import timeit
from typing import Union

import numpy as np
import pandas as pd

from carla.evaluation.distances import get_distances
from carla.evaluation.nearest_neighbours import yNN
from carla.evaluation.process_nans import remove_nans
from carla.evaluation.redundancy import redundancy
from carla.evaluation.success_rate import success_rate
from carla.evaluation.violations import constraint_violation
from carla.models.api import MLModel
from carla.models.catalog import MLModelCatalog
from carla.recourse_methods.api import RecourseMethod
from carla.recourse_methods.processing import get_drop_columns_binary


class Benchmark:
    """
    The benchmarking class contains all measurements.
    It is possible to run only individual evaluation metrics or all via one single call.

    For every given factual, the benchmark object will generate one counterfactual example with
    the given recourse method.

    Parameters
    ----------
    mlmodel: carla.models.MLModel
        Black Box model we want to explain
    recmodel: carla.recourse_methods.RecourseMethod
        Recourse method we want to benchmark
    factuals: pd.DataFrame
        Instances we want to find counterfactuals

    Methods
    -------
    compute_ynn:
        Computes y-Nearest-Neighbours for generated counterfactuals
    compute_average_time:
        Computes average time for generated counterfactual
    compute_distances:
        Calculates the distance measure and returns it as dataframe
    compute_constraint_violation:
        Computes the constraint violation per factual as dataframe
    compute_redundancy:
        Computes redundancy for each counterfactual
    compute_success_rate:
        Computes success rate for the whole recourse method.
    run_benchmark:
        Runs every measurement and returns every value as dict.
    """

    def __init__(
        self,
        mlmodel: Union[MLModel, MLModelCatalog],
        recourse_method: RecourseMethod,
        factuals: pd.DataFrame,
    ) -> None:

        self._mlmodel = mlmodel
        self._recourse_method = recourse_method
        start = timeit.default_timer()
        self._counterfactuals = recourse_method.get_counterfactuals(factuals)
        stop = timeit.default_timer()
        self._timer = stop - start

        # Avoid using scaling and normalizing more than once
        if isinstance(mlmodel, MLModelCatalog):
            self._mlmodel.use_pipeline = False  # type: ignore

        self._factuals = factuals.copy()

        # Normalizing and encoding factual for later use
        self._enc_norm_factuals = recourse_method.encode_normalize_order_factuals(
            factuals, with_target=True
        )

    def compute_ynn(self) -> pd.DataFrame:
        """
        Computes y-Nearest-Neighbours for generated counterfactuals

        Returns
        -------
        pd.DataFrame
        """
        _, counterfactuals_without_nans = remove_nans(
            self._factuals, self._counterfactuals
        )

        if counterfactuals_without_nans.empty:
            ynn = np.nan
        else:
            ynn = yNN(
                counterfactuals_without_nans, self._recourse_method, self._mlmodel, 5
            )

        columns = ["y-Nearest-Neighbours"]

        return pd.DataFrame([[ynn]], columns=columns)

    def compute_average_time(self) -> pd.DataFrame:
        """
        Computes average time for generated counterfactual

        Returns
        -------
        pd.DataFrame
        """

        avg_time = self._timer / self._counterfactuals.shape[0]

        columns = ["Average_Time"]

        return pd.DataFrame([[avg_time]], columns=columns)

    def compute_distances(self) -> pd.DataFrame:
        """
        Calculates the distance measure and returns it as dataframe

        Returns
        -------
        pd.DataFrame
        """
        factual_without_nans, counterfactuals_without_nans = remove_nans(
            self._enc_norm_factuals, self._counterfactuals
        )

        columns = ["Distance_1", "Distance_2", "Distance_3", "Distance_4"]

        if counterfactuals_without_nans.empty:
            return pd.DataFrame(columns=columns)

        if self._mlmodel.encoder.drop is None:
            # To prevent double count of encoded features without drop if_binary
            binary_columns_to_drop = get_drop_columns_binary(
                self._mlmodel.data.categoricals,
                counterfactuals_without_nans.columns.tolist(),
            )
            counterfactuals_without_nans = counterfactuals_without_nans.drop(
                binary_columns_to_drop, axis=1
            )
            factual_without_nans = factual_without_nans.drop(
                binary_columns_to_drop, axis=1
            )

        arr_f = factual_without_nans.to_numpy()
        arr_cf = counterfactuals_without_nans.to_numpy()

        distances = get_distances(arr_f, arr_cf)

        output = pd.DataFrame(distances, columns=columns)

        return output

    def compute_constraint_violation(self) -> pd.DataFrame:
        """
        Computes the constraint violation per factual as dataframe

        Returns
        -------
        pd.Dataframe
        """
        factual_without_nans, counterfactuals_without_nans = remove_nans(
            self._factuals, self._counterfactuals
        )

        if counterfactuals_without_nans.empty:
            violations = []
        else:
            violations = constraint_violation(
                self._mlmodel, counterfactuals_without_nans, factual_without_nans
            )
        columns = ["Constraint_Violation"]

        return pd.DataFrame(violations, columns=columns)

    def compute_redundancy(self) -> pd.DataFrame:
        """
        Computes redundancy for each counterfactual

        Returns
        -------
        pd.Dataframe
        """
        factual_without_nans, counterfactuals_without_nans = remove_nans(
            self._enc_norm_factuals, self._counterfactuals
        )

        if counterfactuals_without_nans.empty:
            redundancies = []
        else:
            redundancies = redundancy(
                factual_without_nans, counterfactuals_without_nans, self._mlmodel
            )

        columns = ["Redundancy"]

        return pd.DataFrame(redundancies, columns=columns)

    def compute_success_rate(self) -> pd.DataFrame:
        """
        Computes success rate for the whole recourse method.

        Returns
        -------
        pd.Dataframe
        """

        rate = success_rate(self._counterfactuals)
        columns = ["Success_Rate"]

        return pd.DataFrame([[rate]], columns=columns)

    def run_benchmark(self) -> pd.DataFrame:
        """
        Runs every measurement and returns every value as dict.

        Returns
        -------
        pd.DataFrame
        """
        pipeline = [
            self.compute_distances(),
            self.compute_constraint_violation(),
            self.compute_redundancy(),
            self.compute_ynn(),
            self.compute_success_rate(),
            self.compute_average_time(),
        ]

        output = pd.concat(pipeline, axis=1)

        return output



from typing import List

import numpy as np


def get_distances(factual: np.ndarray, counterfactual: np.ndarray) -> List[List[float]]:
    """
    Computes distances 1 to 4.
    All features have to be in the same order (without target label).

    Parameters
    ----------
    factual: np.ndarray
        Normalized and encoded array with factual data.
        Shape: NxM
    counterfactual: np.ndarray
        Normalized and encoded array with counterfactual data
        Shape: NxM

    Returns
    -------
    list: distances 1 to 4
    """
    if factual.shape != counterfactual.shape:
        raise ValueError("Shapes of factual and counterfactual have to be the same")
    if len(factual.shape) != 2:
        raise ValueError(
            "Shapes of factual and counterfactual have to be 2-dimensional"
        )

    # get difference between original and counterfactual
    delta = get_delta(factual, counterfactual)

    d1 = d1_distance(delta)
    d2 = d2_distance(delta)
    d3 = d3_distance(delta)
    d4 = d4_distance(delta)

    return [[d1[i], d2[i], d3[i], d4[i]] for i in range(len(d1))]


def d1_distance(delta: np.ndarray) -> List[float]:
    """
    Computes D1 distance

    Parameters
    ----------
    delta: np.ndarray
        Difference between factual and counterfactual

    Returns
    -------
    List[float]
    """
    # compute elements which are greater than 0
    return np.sum(delta != 0, axis=1, dtype=np.float).tolist()


def d2_distance(delta: np.ndarray) -> List[float]:
    """
    Computes D2 distance

    Parameters
    ----------
    delta: np.ndarray
        Difference between factual and counterfactual

    Returns
    -------
    List[float]
    """

    return np.sum(np.abs(delta), axis=1, dtype=np.float).tolist()


def d3_distance(delta: np.ndarray) -> List[float]:
    """
    Computes D3 distance

    Parameters
    ----------
    delta: np.ndarray
        Difference between factual and counterfactual

    Returns
    -------
    List[float]
    """
    return np.sum(np.square(np.abs(delta)), axis=1, dtype=np.float).tolist()


def d4_distance(delta: np.ndarray) -> List[float]:
    """
    Computes D4 distance

    Parameters
    ----------
    delta: np.ndarray
        Difference between factual and counterfactual

    Returns
    -------
    List[float]
    """
    return np.max(np.abs(delta), axis=1).tolist()


def get_delta(instance: np.ndarray, cf: np.ndarray) -> np.ndarray:
    """
    Compute difference between original instance and counterfactual

    Parameters
    ----------
    instance: np.ndarray
        Normalized and encoded array with factual data.
        Shape: NxM
    cf: : np.ndarray
        Normalized and encoded array with counterfactual data.
        Shape: NxM

    Returns
    -------
    np.ndarray
    """
    return cf - instance



from typing import Dict, List, Tuple, Union

import numpy as np
import pandas as pd
import torch
from numpy import linalg as LA

from carla import log
from carla.models.api import MLModel
from carla.recourse_methods.api import RecourseMethod
from carla.recourse_methods.autoencoder import (
    VariationalAutoencoder,
    train_variational_autoencoder,
)
from carla.recourse_methods.processing import (
    check_counterfactuals,
    merge_default_parameters,
    reconstruct_encoding_constraints,
)


class CCHVAE(RecourseMethod):
    """
    Implementation of CCHVAE [1]_

    Parameters
    ----------
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Dictionary containing hyperparameters. See Notes below to see its content.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Notes
    -----
    - Hyperparams
        Hyperparameter contains important information for the recourse method to initialize.
        Please make sure to pass all values as dict with the following keys.

        * "data_name": str
            name of the dataset
        * "n_search_samples": int, default: 300
            Number of generated candidate counterfactuals.
        * "p_norm": {1, 2}
            Defines L_p norm for distance calculation.
        * "step": float, default: 0.1
            Step size for each generated candidate counterfactual.
        * "max_iter": int, default: 1000
            Number of iterations per factual instance.
        * "clamp": bool, default: True
            Feature values will be clamped between 0 and 1
        * "binary_cat_features": bool, default: True
            If true, the encoding of x is done by drop_if_binary.
        * "vae_params": Dict
            With parameter for VAE.

            + "layers": list
                List with number of neurons per layer.
            + "train": bool, default: True
                Decides if a new Autoencoder will be learned.
            + "lambda_reg": float, default: 1e-6
                Hyperparameter for variational autoencoder.
            + "epochs": int, default: 5
                Number of epochs to train VAE
            + "lr": float, default: 1e-3
                Learning rate for VAE training
            + "batch_size": int, default: 32
                Batch-size for VAE training

    .. [1] Pawelczyk, Martin, Klaus Broelemann and Gjergji Kasneci. “Learning Model-Agnostic Counterfactual Explanations
          for Tabular Data.” Proceedings of The Web Conference 2020 (2020): n. pag..
    """

    _DEFAULT_HYPERPARAMS = {
        "data_name": None,
        "n_search_samples": 300,
        "p_norm": 1,
        "step": 0.1,
        "max_iter": 1000,
        "clamp": True,
        "binary_cat_features": True,
        "vae_params": {
            "layers": None,
            "train": True,
            "lambda_reg": 1e-6,
            "epochs": 5,
            "lr": 1e-3,
            "batch_size": 32,
        },
    }

    def __init__(self, mlmodel: MLModel, hyperparams: Dict) -> None:
        super().__init__(mlmodel)
        self._params = merge_default_parameters(hyperparams, self._DEFAULT_HYPERPARAMS)

        df_enc_norm_data = self.encode_normalize_order_factuals(
            self._mlmodel.data.raw, with_target=True
        )

        self._n_search_samples = self._params["n_search_samples"]
        self._p_norm = self._params["p_norm"]
        self._step = self._params["step"]
        self._max_iter = self._params["max_iter"]
        self._clamp = self._params["clamp"]

        vae_params = self._params["vae_params"]
        self._generative_model = self._load_vae(
            df_enc_norm_data, vae_params, self._mlmodel, self._params["data_name"]
        )

    def _load_vae(
        self, data: pd.DataFrame, vae_params: Dict, mlmodel: MLModel, data_name: str
    ) -> VariationalAutoencoder:
        generative_model = VariationalAutoencoder(
            data_name,
            vae_params["layers"],
        )

        if vae_params["train"]:
            generative_model = train_variational_autoencoder(
                generative_model,
                mlmodel.data,
                mlmodel.scaler,
                mlmodel.encoder,
                mlmodel.feature_input_order,
                lambda_reg=vae_params["lambda_reg"],
                epochs=vae_params["epochs"],
                lr=vae_params["lr"],
                batch_size=vae_params["batch_size"],
            )
        else:
            try:
                generative_model.load(data.shape[1] - 1)
            except FileNotFoundError as exc:
                raise FileNotFoundError(
                    "Loading of Autoencoder failed. {}".format(str(exc))
                )

        return generative_model

    def _hyper_sphere_coordindates(
        self, instance, high: int, low: int
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        :param n_search_samples: int > 0
        :param instance: numpy input point array
        :param high: float>= 0, h>l; upper bound
        :param low: float>= 0, l<h; lower bound
        :param p: float>= 1; norm
        :return: candidate counterfactuals & distances
        """
        delta_instance = np.random.randn(self._n_search_samples, instance.shape[1])
        dist = (
            np.random.rand(self._n_search_samples) * (high - low) + low
        )  # length range [l, h)
        norm_p = LA.norm(delta_instance, ord=self._p_norm, axis=1)
        d_norm = np.divide(dist, norm_p).reshape(-1, 1)  # rescale/normalize factor
        delta_instance = np.multiply(delta_instance, d_norm)
        candidate_counterfactuals = instance + delta_instance
        return candidate_counterfactuals, dist

    def _counterfactual_search(
        self, step: int, factual: torch.Tensor, cat_features_indices: List
    ) -> pd.DataFrame:
        device = "cuda" if torch.cuda.is_available() else "cpu"

        # init step size for growing the sphere
        low = 0
        high = step
        # counter
        count = 0
        counter_step = 1

        torch_fact = torch.from_numpy(factual).to(device)

        # get predicted label of instance
        instance_label = np.argmax(
            self._mlmodel.predict_proba(torch_fact.float()).cpu().detach().numpy(),
            axis=1,
        )

        # vectorize z
        z = self._generative_model.encode(torch_fact.float())[0].cpu().detach().numpy()
        z_rep = np.repeat(z.reshape(1, -1), self._n_search_samples, axis=0)

        candidate_dist: List = []
        x_ce: Union[np.ndarray, torch.Tensor] = np.array([])
        while count <= self._max_iter or len(candidate_dist) <= 0:
            count = count + counter_step
            if count > self._max_iter:
                log.debug("No counterfactual example found")
                return x_ce[0]

            # STEP 1 -- SAMPLE POINTS on hyper sphere around instance
            latent_neighbourhood, _ = self._hyper_sphere_coordindates(z_rep, high, low)
            torch_latent_neighbourhood = (
                torch.from_numpy(latent_neighbourhood).to(device).float()
            )
            x_ce = self._generative_model.decode(torch_latent_neighbourhood)[0]
            x_ce = reconstruct_encoding_constraints(
                x_ce, cat_features_indices, self._params["binary_cat_features"]
            )
            x_ce = x_ce.detach().cpu().numpy()
            x_ce = x_ce.clip(0, 1) if self._clamp else x_ce

            # STEP 2 -- COMPUTE l1 & l2 norms
            if self._p_norm == 1:
                distances = np.abs((x_ce - torch_fact.cpu().detach().numpy())).sum(
                    axis=1
                )
            elif self._p_norm == 2:
                distances = LA.norm(x_ce - torch_fact.cpu().detach().numpy(), axis=1)
            else:
                raise ValueError("Possible values for p_norm are 1 or 2")

            # counterfactual labels
            y_candidate = np.argmax(
                self._mlmodel.predict_proba(torch.from_numpy(x_ce).float())
                .cpu()
                .detach()
                .numpy(),
                axis=1,
            )
            indeces = np.where(y_candidate != instance_label)
            candidate_counterfactuals = x_ce[indeces]
            candidate_dist = distances[indeces]
            # no candidate found & push search range outside
            if len(candidate_dist) == 0:
                low = high
                high = low + step
            elif len(candidate_dist) > 0:
                # certain candidates generated
                min_index = np.argmin(candidate_dist)
                log.debug("Counterfactual example found")
                return candidate_counterfactuals[min_index]

    def get_counterfactuals(self, factuals: pd.DataFrame) -> pd.DataFrame:
        df_enc_norm_fact = self.encode_normalize_order_factuals(factuals)

        encoded_feature_names = self._mlmodel.encoder.get_feature_names(
            self._mlmodel.data.categoricals
        )
        cat_features_indices = [
            df_enc_norm_fact.columns.get_loc(feature)
            for feature in encoded_feature_names
        ]

        df_cfs = df_enc_norm_fact.apply(
            lambda x: self._counterfactual_search(
                self._step, x.reshape((1, -1)), cat_features_indices
            ),
            raw=True,
            axis=1,
        )

        df_cfs = check_counterfactuals(self._mlmodel, df_cfs)

        return df_cfs



# import helpers
import copy

import numpy as np

# import Dijkstra's shortest path algorithm
from scipy.sparse import csgraph, csr_matrix

# import graph building methods
from sklearn.neighbors import kneighbors_graph, radius_neighbors_graph


def graph_search(
    data,
    index,
    keys_immutable,
    model,
    n_neighbors=50,
    p_norm=2,
    mode="knn",
    frac=0.4,
    radius=0.25,
):
    # This one implements the FACE method from
    # Rafael Poyiadzi et al (2020), "FACE: Feasible and Actionable Counterfactual Explanations",
    # Conference on AI, Ethics & Accountability (AIES, 2020)
    """
    :param data: df
    :param n_neighbors: int > 0; number of neighbors when constructing knn graph
    :param step: float > 0; step_size for growing spheres
    :param mode: str; either 'knn' or 'epsilon'
    :param model: classification model (either tf keras, pytorch or sklearn)
    :param p_norm: float=>1; denotes the norm (classical: 1 or 2)
    :param frac: float 0 < number =< 1; fraction of data for which we compute the graph; if frac = 1, and data set large, then compute long
    :param keys_immutable: list; list of input names that may not be searched over
    :param radius: float > 0; parameter for epsilon density graph
    :return: candidate_counterfactual_star: np array (min. cost counterfactual explanation)
    """
    # Choose a subset of data for computational efficiency
    data = choose_random_subset(data, frac, index)

    # ADD CONSTRAINTS by immutable inputs into adjacency matrix
    # if element in adjacency matrix 0, then it cannot be reached
    # this ensures that paths only take same sex / same race / ... etc. routes
    for i in range(len(keys_immutable)):
        immutable_constraint_matrix1, immutable_constraint_matrix2 = build_constraints(
            data, i, keys_immutable
        )

    # POSITIVE PREDICTIONS
    y_predicted = model.predict_proba(data.values)
    y_predicted = np.argmax(y_predicted, axis=1)
    y_positive_indeces = np.where(y_predicted == 1)

    if mode == "knn":
        boundary = 3  # chosen in ad-hoc fashion
        median = n_neighbors
        is_knn = True

    elif mode == "epsilon":
        boundary = 0.10  # chosen in ad-hoc fashion
        median = radius
        is_knn = False
    else:
        raise ValueError("Only possible values for mode are knn and epsilon")

    neighbors_list = [
        median - boundary,
        median,
        median + boundary,
    ]

    # obtain candidate targets (CT); two conditions need to be met:
    # (1) CT needs to be predicted positively & (2) CT needs to have certain "density"
    # for knn I interpret 'certain density' as sufficient number of neighbours
    candidate_counterfactuals = []
    for n in neighbors_list:
        neighbor_candidates = find_counterfactuals(
            candidate_counterfactuals,
            data,
            immutable_constraint_matrix1,
            immutable_constraint_matrix2,
            index,
            n,
            y_positive_indeces,
            is_knn=is_knn,
        )
        candidate_counterfactuals += neighbor_candidates

    candidate_counterfactual_star = np.array(candidate_counterfactuals)

    # STEP 4 -- COMPUTE DISTANCES between x^F and candidate x^CF; else return NaN
    if candidate_counterfactual_star.size == 0:
        candidate_counterfactual_star = np.empty(
            data.values.shape[1],
        )
        candidate_counterfactual_star[:] = np.nan

        return candidate_counterfactual_star

    if p_norm == 1:
        c_dist = np.abs((data.values[index] - candidate_counterfactual_star)).sum(
            axis=1
        )
    elif p_norm == 2:
        c_dist = np.square((data.values[index] - candidate_counterfactuals)).sum(axis=1)
    else:
        raise ValueError("Distance not defined yet. Choose p_norm to be 1 or 2")

    min_index = np.argmin(c_dist)
    candidate_counterfactual_star = candidate_counterfactual_star[min_index]

    return candidate_counterfactual_star


def choose_random_subset(data, frac, index):
    """
    Choose a subset of data for computational efficiency

    Parameters
    ----------
    data : pd.DataFrame
    frac: float 0 < number =< 1
        fraction of data for which we compute the graph; if frac = 1, and data set large, then compute long
    index: int

    Returns
    -------
    pd.DataFrame
    """
    number_samples = np.int(np.rint(frac * data.values.shape[0]))
    list_to_choose = (
        np.arange(0, index).tolist()
        + np.arange(index + 1, data.values.shape[0]).tolist()
    )
    chosen_indeces = np.random.choice(
        list_to_choose,
        replace=False,
        size=number_samples,
    )
    chosen_indeces = [
        index
    ] + chosen_indeces.tolist()  # make sure sample under consideration included
    data = data.iloc[chosen_indeces]
    data = data.sort_index()
    return data


def build_constraints(data, i, keys_immutable, epsilon=0.5):
    """

    Parameters
    ----------
    data: pd.DataFrame
    i : int
        Position of immutable key
    keys_immutable: list[str]
        Immutable feature
    epsilon: int

    Returns
    -------
    np.ndarray, np.ndarray
    """
    immutable_constraint_matrix = np.outer(
        data[keys_immutable[i]].values + epsilon,
        data[keys_immutable[i]].values + epsilon,
    )
    immutable_constraint_matrix1 = immutable_constraint_matrix / ((1 + epsilon) ** 2)
    immutable_constraint_matrix1 = ((immutable_constraint_matrix1 == 1) * 1).astype(
        float
    )
    immutable_constraint_matrix2 = immutable_constraint_matrix / (epsilon ** 2)
    immutable_constraint_matrix2 = ((immutable_constraint_matrix2 == 1) * 1).astype(
        float
    )
    return immutable_constraint_matrix1, immutable_constraint_matrix2


def find_counterfactuals(
    candidates,
    data,
    immutable_constraint_matrix1,
    immutable_constraint_matrix2,
    index,
    n,
    y_positive_indeces,
    is_knn,
):
    """
    Steps 1 to 3 of the FACE algorithm

    Parameters
    ----------
    candidate_counterfactuals_star: list
    data: pd.DataFrame
    immutable_constraint_matrix1: np.ndarray
    immutable_constraint_matrix2: np.ndarray
    index: int
    n: int
    y_positive_indeces: int
    is_knn: bool

    Returns
    -------
    list
    """
    candidate_counterfactuals_star = copy.deepcopy(candidates)
    # STEP 1 -- BUILD NETWORK GRAPH
    graph = build_graph(
        data, immutable_constraint_matrix1, immutable_constraint_matrix2, is_knn, n
    )
    # STEP 2 -- APPLY SHORTEST PATH ALGORITHM  ## indeces=index (corresponds to x^F)
    distances, min_distance = shortest_path(graph, index)
    # STEP 3 -- FIND COUNTERFACTUALS
    # minimum distance candidate counterfactuals
    candidate_min_distances = [
        min_distance,
        min_distance + 1,
        min_distance + 2,
        min_distance + 3,
    ]
    min_distance_indeces = np.array([0])
    for min_dist in candidate_min_distances:
        min_distance_indeces = np.c_[
            min_distance_indeces, np.array(np.where(distances == min_dist))
        ]
    min_distance_indeces = np.delete(min_distance_indeces, 0)
    indeces_counterfactuals = np.intersect1d(
        np.array(y_positive_indeces), np.array(min_distance_indeces)
    )
    for i in range(indeces_counterfactuals.shape[0]):
        candidate_counterfactuals_star.append(data.values[indeces_counterfactuals[i]])

    return candidate_counterfactuals_star


def shortest_path(graph, index):
    """
    Uses dijkstras shortest path

    Parameters
    ----------
    graph: CSR matrix
    index: int

    Returns
    -------
    np.ndarray, float
    """
    distances = csgraph.dijkstra(
        csgraph=graph, directed=False, indices=index, return_predecessors=False
    )
    distances[index] = np.inf  # avoid min. distance to be x^F itself
    min_distance = distances.min()
    return distances, min_distance


def build_graph(
    data, immutable_constraint_matrix1, immutable_constraint_matrix2, is_knn, n
):
    """

    Parameters
    ----------
    data: pd.DataFrame
    immutable_constraint_matrix1: np.ndarray
    immutable_constraint_matrix2: np.ndarray
    is_knn: bool
    n: int

    Returns
    -------
    CSR matrix
    """
    if is_knn:
        graph = kneighbors_graph(data.values, n_neighbors=n, n_jobs=-1)
    else:
        graph = radius_neighbors_graph(data.values, radius=n, n_jobs=-1)
    adjacency_matrix = graph.toarray()
    adjacency_matrix = np.multiply(
        adjacency_matrix,
        immutable_constraint_matrix1,
        immutable_constraint_matrix2,
    )  # element wise multiplication
    graph = csr_matrix(adjacency_matrix)
    return graph



from typing import Any, Dict

import dice_ml
import pandas as pd

from carla.models.api import MLModel

from ...api import RecourseMethod
from ...processing import merge_default_parameters


class Dice(RecourseMethod):
    """
    Implementation of Dice from Mothilal et.al. [1]_.

    Parameters
    ----------
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Dictionary containing hyperparameters. See notes below for its contents.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Notes
    -----
    - Hyperparams
        Hyperparameter contains important information for the recourse method to initialize.
        Please make sure to pass all values as dict with the following keys.

        * "num": int, default: 1
            Number of counterfactuals per factual to generate
        * "desired_class": int, default: 1
            Given a binary class label, the desired class a counterfactual should have (e.g., 0 or 1)
        * "posthoc_sparsity_param": float, default: 0.1
            Fraction of post-hoc preprocessing steps.
    - Restrictions:
        *   Only the model agnostic approach (backend: sklearn) is used in our implementation.
        *   ML model needs to have a transformation pipeline for normalization, encoding and feature order.
            See pipelining at carla/models/catalog/catalog.py for an example ML model class implementation

    .. [1] R. K. Mothilal, Amit Sharma, and Chenhao Tan. 2020. Explaining machine learning classifiers
            through diverse counterfactual explanations
    """

    _DEFAULT_HYPERPARAMS = {"num": 1, "desired_class": 1, "posthoc_sparsity_param": 0.1}

    def __init__(self, mlmodel: MLModel, hyperparams: Dict[str, Any]) -> None:
        super().__init__(mlmodel)
        self._continous = mlmodel.data.continous
        self._categoricals = mlmodel.data.categoricals
        self._target = mlmodel.data.target

        checked_hyperparams = merge_default_parameters(
            hyperparams, self._DEFAULT_HYPERPARAMS
        )
        # Prepare data for dice data structure
        self._dice_data = dice_ml.Data(
            dataframe=mlmodel.data.raw,
            continuous_features=self._continous,
            outcome_name=self._target,
        )

        self._dice_model = dice_ml.Model(model=mlmodel, backend="sklearn")

        self._dice = dice_ml.Dice(self._dice_data, self._dice_model, method="random")
        self._num = checked_hyperparams["num"]
        self._desired_class = checked_hyperparams["desired_class"]
        self._post_hoc_sparsity_param = checked_hyperparams["posthoc_sparsity_param"]

        # Need scaler and encoder for get_counterfactual output
        self._scaler = mlmodel.scaler
        self._encoder = mlmodel.encoder
        self._feature_order = mlmodel.feature_input_order

    def get_counterfactuals(self, factuals: pd.DataFrame) -> pd.DataFrame:
        # Prepare factuals
        querry_instances = factuals.copy()

        # check if querry_instances are not empty
        if not querry_instances.shape[0] > 0:
            raise ValueError("Factuals should not be empty")

        # Generate counterfactuals
        dice_exp = self._dice.generate_counterfactuals(
            querry_instances,
            total_CFs=self._num,
            desired_class=self._desired_class,
            posthoc_sparsity_param=self._post_hoc_sparsity_param,
        )

        list_cfs = dice_exp.cf_examples_list
        df_cfs = pd.concat([cf.final_cfs_df for cf in list_cfs], ignore_index=True)
        df_cfs[self._continous] = self._scaler.transform(df_cfs[self._continous])
        encoded_features = self._encoder.get_feature_names(self._categoricals)
        df_cfs[encoded_features] = self._encoder.transform(df_cfs[self._categoricals])
        df_cfs = df_cfs[self._feature_order + [self._target]]

        return df_cfs



from __future__ import division, print_function

import time

import torch.utils.data
from numpy.random import normal
from torchvision.utils import make_grid

from carla.recourse_methods.catalog.clue.library.clue_ml.src.utils import *


def train_VAE(
    net,
    name,
    batch_size,
    nb_epochs,
    trainset,
    valset,
    cuda,
    flat_ims=False,
    train_plot=False,
    Nclass=None,
    early_stop=None,
    script_mode=False,
):

    models_dir = name
    results_dir = name + "_results"

    if cuda:
        trainloader = torch.utils.data.DataLoader(
            trainset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=True,
            num_workers=0,
        )
        valloader = torch.utils.data.DataLoader(
            valset, batch_size=batch_size, shuffle=False, pin_memory=True, num_workers=0
        )

    else:
        trainloader = torch.utils.data.DataLoader(
            trainset,
            batch_size=batch_size,
            shuffle=True,
            pin_memory=False,
            num_workers=0,
        )
        valloader = torch.utils.data.DataLoader(
            valset,
            batch_size=batch_size,
            shuffle=False,
            pin_memory=False,
            num_workers=0,
        )

    ## ---------------------------------------------------------------------------------------------------------------------
    # net dims
    log.info("\nNetwork:")

    epoch = 0

    ## ---------------------------------------------------------------------------------------------------------------------
    # train
    log.info("\nTrain:")

    log.info("init cost variables:")
    vlb_train = np.zeros(nb_epochs)
    vlb_dev = np.zeros(nb_epochs)
    best_vlb = -np.inf
    best_vlb_train = -np.inf
    best_epoch = 0

    nb_its_dev = 1

    tic0 = time.time()
    for i in range(epoch, nb_epochs):
        net.set_mode_train(True)

        tic = time.time()
        nb_samples = 0
        for x, y in trainloader:

            if flat_ims:
                x = x.view(x.shape[0], -1)
            if Nclass is not None:
                y_oh = torch_onehot(y, Nclass).type(x.type())
                x = torch.cat([x, y_oh], 1)

            cost, _ = net.fit(x)

            vlb_train[i] += cost * len(x)
            nb_samples += len(x)

        vlb_train[i] /= nb_samples

        toc = time.time()

        # ---- print
        log.info("it %d/%d, vlb %f, " % (i, nb_epochs, vlb_train[i]))
        log.info("time: %f seconds\n" % (toc - tic))
        net.update_lr(i)

        if vlb_train[i] > best_vlb_train:
            best_vlb_train = vlb_train[i]

        # ---- dev
        if i % nb_its_dev == 0:
            nb_samples = 0
            for j, (x, y) in enumerate(valloader):

                if flat_ims:
                    x = x.view(x.shape[0], -1)
                if Nclass is not None:
                    y_oh = torch_onehot(y, Nclass).type(x.type())
                    x = torch.cat([x, y_oh], 1)

                cost, _ = net.eval(x)

                vlb_dev[i] += cost * len(x)
                nb_samples += len(x)

            vlb_dev[i] /= nb_samples

            log.info("vlb %f (%f)\n" % (vlb_dev[i], best_vlb))

            if train_plot:
                zz = net.recongnition(x).sample()
                o = net.regenerate(zz)
                try:
                    o = o.cpu()
                except:
                    o = o.loc.cpu()
                if len(x.shape) == 2:
                    side = int(np.sqrt(x.shape[1]))
                    x = x.view(-1, 1, side, side).data
                    o = o.view(-1, 1, side, side).data

                import matplotlib.pyplot as plt

                plt.figure()
                dd = make_grid(torch.cat([x[:10], o[:10]]), nrow=10).numpy()
                plt.imshow(np.transpose(dd, (1, 2, 0)), interpolation="nearest")
                if script_mode:
                    plt.savefig(results_dir + "/rec%d.png" % i)
                else:
                    plt.show()

                z_sample = normal(loc=0.0, scale=1.0, size=(36, net.latent_dim))
                x_rec = net.regenerate(z_sample)
                try:
                    x_rec = x_rec.cpu()
                except:
                    x_rec = x_rec.loc.cpu()
                if len(x_rec.shape) == 2:
                    side = int(np.sqrt(x_rec.shape[1]))
                    x_rec = x_rec.view(-1, 1, side, side)
                plt.figure()
                dd = make_grid(x_rec, nrow=6).numpy()
                plt.imshow(np.transpose(dd, (1, 2, 0)), interpolation="nearest")
                if script_mode:
                    plt.savefig(results_dir + "/sample%d.png" % i)
                else:
                    plt.show()

        if vlb_dev[i] > best_vlb:
            best_vlb = vlb_dev[i]
            best_epoch = i
            net.save(os.path.join(models_dir, "theta_best.dat"))

        if early_stop is not None and (i - best_epoch) > early_stop:
            break

    net.save(models_dir + "/theta_last.dat")
    toc0 = time.time()
    runtime_per_it = (toc0 - tic0) / float(nb_epochs)
    log.info("average time: %f seconds\n" % runtime_per_it)

    ## ---------------------------------------------------------------------------------------------------------------------
    # results
    log.info("\nRESULTS:")
    nb_parameters = net.get_nb_parameters()
    best_cost_dev = best_vlb
    best_cost_train = best_vlb_train

    log.info("best_vlb_dev: %f" % best_cost_dev)
    log.info("best_vlb_train: %f" % best_cost_train)
    log.info("nb_parameters: %d (%s)\n" % (nb_parameters, humansize(nb_parameters)))

    ## ---------------------------------------------------------------------------------------------------------------------
    return vlb_train, vlb_dev



# flake8: noqa
# isort:skip
import logging.config
import os
import pathlib

import yaml

lib_path = pathlib.Path(__file__).parent.resolve()
with open(os.path.join(lib_path, "logging.yaml"), "r") as f:
    config = yaml.safe_load(f.read())
    logging.config.dictConfig(config)

log = logging.getLogger(__name__)

from ._version import __version__
from .data import Data, DataCatalog
from .evaluation import Benchmark
from .models import MLModel, MLModelCatalog
from .recourse_methods import RecourseMethod


def get_logger(logger: str):
    return logging.getLogger(logger)



from __future__ import division, print_function

import torch.nn.functional as F
from torch.optim import Adam

from carla import log
from carla.recourse_methods.catalog.clue.library.clue_ml.src.probability import (
    decompose_entropy_cat,
    decompose_std_gauss,
)
from carla.recourse_methods.catalog.clue.library.clue_ml.src.utils import *
from carla.recourse_methods.catalog.clue.library.clue_ml.src.utils import Ln_distance

"""
Here we conduct the search for counterfactual explanations using CLUE.
We downloaded the CLUE method and its helpers from openreview:
have a look at: https://openreview.net/forum?id=XSLF1XFq5h for more details
"""


def vae_gradient_search(
    instance,
    model,
    VAE,
    lr=0.5,
    prediction_similarity_weight=0,
    aleatoric_weight=0,
    epistemic_weight=0,
    uncertainty_weight=1,
    lambda_param=4,
    prior_weight=0,
    latent_L2_weight=0,
    min_steps=3,
    max_steps=400,
    n_early_stop=3,
):
    # Most default hyper parameters are chosen to match those from the paper
    # Data set specific hyper parameters are set as follows:
    # lr: set as the average of data set specific learning rates
    # lambda: set as the average of data set specific learning rates
    # it's likely that 'lr' and 'lambda' require data set specific tuning - ask Javier how he went about this

    """
    :param instance: np array
    :param keys_mutable: list;
    :param keys_immutable:list;
    :param continuous_cols: list;
    :param binary_cols: list;
    :param model: pretrained tf model
    :param lr: 0< float <1; learning rate
    :param aleatoric_weight: 0< float <1
    :param epistemic_weight: 0< float <1
    :param uncertainty_weight: 0< float <1
    :param prediction_similarity_weight: 0< float <1
    :param lambda_param: 0 < float
    :param min_steps: int > 0
    :param max_steps: int > 0
    :param n_early_stop: int > 0
    :return: counterfactual instance: np array
    """  #

    torch.cuda.empty_cache()

    dist = Ln_distance(n=1, dim=(1))
    instance = instance.reshape(1, -1)
    x_dim = instance.reshape(instance.shape[0], -1).shape[1]
    # Both weights are set to 1?
    distance_weight = lambda_param / x_dim
    desired_preds = model.predict_proba(instance)

    CLUE_explainer = CLUE(
        VAE,
        model,
        instance,
        uncertainty_weight=uncertainty_weight,
        aleatoric_weight=aleatoric_weight,
        epistemic_weight=epistemic_weight,
        prior_weight=prior_weight,
        distance_weight=distance_weight,
        latent_L2_weight=latent_L2_weight,
        prediction_similarity_weight=prediction_similarity_weight,
        lr=lr,
        desired_preds=desired_preds,
        cond_mask=None,
        distance_metric=dist,
        norm_MNIST=False,
        prob_BNN=False,
        flatten_BNN=False,
        regression=False,
        cuda=False,
    )

    torch.autograd.set_detect_anomaly(False)

    # clue_instance.optimizer = SGD(self.trainable_params, lr=lr, momentum=0.5, nesterov=True)
    (
        z_vec,
        counterfactual,
        uncertainty_vec,
        epistemic_vec,
        aleatoric_vec,
        cost_vec,
        dist_vec,
    ) = CLUE_explainer.optimise(
        min_steps=min_steps, max_steps=max_steps, n_early_stop=n_early_stop
    )

    """
    Check if explanation indeed produced counterfactual:
    We accept, if CLUE produced any counterfactual within these max_steps.
    Otherwise we would very often find no counterfactual.
    This 'issue' is also known by the authors who show in the Appendix that CLUE
    produced counterfactuals not more than 20% percent of the time. E.g. running optimization to end
    leads to 16% of successfuly generated counterfactuals for their LSAT data set.
    """

    cf_preds = model.predict_proba(counterfactual[:, 0, :])
    indeces_counterfactual = np.where(
        np.argmax(cf_preds, axis=1) != np.argmax(desired_preds, axis=1)
    )[0]

    # Return np.nan if none found & otherwise return l_1 cost counterfactual
    if not len(indeces_counterfactual):
        counterfactual = counterfactual[max_steps, :, :]
        counterfactual[:] = np.nan
    else:
        distance = np.abs(instance - counterfactual[indeces_counterfactual, :, :]).sum(
            axis=2
        )
        index_min = indeces_counterfactual[np.argmin(distance)]
        counterfactual = counterfactual[index_min, :, :]

    return counterfactual.reshape(-1)


class CLUE(BaseNet):
    """CLUE authors: This will be a general class for CLUE, etc.
    A propper optimiser will be used instead of my manually designed one."""

    def __init__(
        self,
        VAE,
        BNN,
        original_x,
        uncertainty_weight,
        aleatoric_weight,
        epistemic_weight,
        prior_weight,
        distance_weight,
        latent_L2_weight,
        prediction_similarity_weight,
        lr,
        desired_preds=None,
        cond_mask=None,
        distance_metric=None,
        z_init=None,
        norm_MNIST=False,
        flatten_BNN=False,
        regression=False,
        prob_BNN=True,
        cuda=False,
    ):

        """Option specification:
        MNIST: boolean, specifies whether to apply normalisation to VAE outputs before being passed on to the BNN"""
        # Load models
        self.VAE = VAE
        self.BNN = BNN

        # Objective function definition
        self.uncertainty_weight = uncertainty_weight
        self.aleatoric_weight = aleatoric_weight
        self.epistemic_weight = epistemic_weight
        self.prior_weight = prior_weight
        self.distance_weight = distance_weight
        self.distance_metric = distance_metric

        self.latent_L2_weight = latent_L2_weight
        self.prediction_similarity_weight = prediction_similarity_weight
        self.desired_preds = desired_preds

        # Other CLUE parameters
        self.regression = regression
        self.flatten_BNN = flatten_BNN
        self.norm_MNIST = norm_MNIST
        self.original_x = torch.Tensor(original_x)

        self.prob_BNN = prob_BNN
        # self.cuda = cuda
        self.cuda = torch.cuda.is_available()
        if self.cuda:
            self.original_x = self.original_x.cuda()
            # self.z_init = self.z_init.cuda()
            if self.desired_preds is not None:
                # self.desired_preds = self.desired_preds.cuda()
                self.desired_preds = torch.from_numpy(self.desired_preds).cuda()
        else:
            self.desired_preds = torch.from_numpy(self.desired_preds)

        self.cond_mask = cond_mask

        # Trainable params
        self.trainable_params = list()

        if self.VAE is None:  # this will be for ablation test: sensitivity analisys
            self.trainable_params.append(nn.Parameter(original_x))
        else:
            self.z_dim = VAE.latent_dim
            if z_init is not None:
                self.z_init = torch.Tensor(z_init)
                if cuda:
                    self.z_init = self.z_init.cuda()
                self.z = nn.Parameter(self.z_init)
                self.trainable_params.append(self.z)
            else:
                self.z_init = (
                    torch.zeros(self.z_dim).unsqueeze(0).repeat(original_x.shape[0], 1)
                )
                if self.cuda:
                    self.z_init = self.z_init.cuda()
                self.z = nn.Parameter(self.z_init)
                self.trainable_params.append(self.z)

        # Optimiser
        self.optimizer = Adam(self.trainable_params, lr=lr)

    def randomise_z_init(self, std):
        # assert (self.z.data == self.z_init).all()
        eps = torch.randn(self.z.shape).type(self.z.type())
        self.z.data = std * eps + self.z_init
        return None

    def pred_dist(self, preds):
        # We dont implement for now as we could just use VAEAC with class conditioning
        assert self.desired_preds is not None

        if self.regression:
            dist = F.mse_loss(preds, self.desired_preds, reduction="none")
        else:

            if len(self.desired_preds.shape) == 1 or self.desired_preds.shape[1] == 1:
                dist = F.nll_loss(preds, self.desired_preds, reduction="none")
            else:  # Soft cross entropy loss
                dist = -(torch.log(preds) * self.desired_preds).sum(dim=1)
        return dist

    def uncertainty_from_z(self):
        # We dont use unflatten option because BNNs always take flattened input and unflatten doesnt support grad
        x = self.VAE.regenerate(self.z, grad=True)

        if self.flatten_BNN:
            to_BNN = x.view(x.shape[0], -1)
        else:
            to_BNN = x

        if self.norm_MNIST:
            to_BNN = MNIST_mean_std_norm(to_BNN)

        if self.prob_BNN:
            if self.regression:
                mu_vec, std_vec = self.BNN.sample_predict(to_BNN, Nsamples=0, grad=True)
                (
                    total_uncertainty,
                    aleatoric_uncertainty,
                    epistemic_uncertainty,
                ) = decompose_std_gauss(mu_vec, std_vec)
                preds = mu_vec.mean(dim=0)
            else:
                probs = self.BNN.sample_predict(to_BNN, Nsamples=0, grad=True)
                (
                    total_uncertainty,
                    aleatoric_uncertainty,
                    epistemic_uncertainty,
                ) = decompose_entropy_cat(probs)
                preds = probs.mean(dim=0)
        else:
            if self.regression:
                mu, std = self.BNN.predict(to_BNN, grad=True)
                total_uncertainty = std.squeeze(1)
                aleatoric_uncertainty = total_uncertainty
                epistemic_uncertainty = total_uncertainty * 0
                preds = mu
            else:
                probs = self.BNN.predict_proba(to_BNN)
                total_uncertainty = -(probs * torch.log(probs + 1e-10)).sum(
                    dim=1, keepdim=False
                )
                aleatoric_uncertainty = total_uncertainty
                epistemic_uncertainty = total_uncertainty * 0
                preds = probs

        return total_uncertainty, aleatoric_uncertainty, epistemic_uncertainty, x, preds

    def get_objective(
        self, x, total_uncertainty, aleatoric_uncertainty, epistemic_uncertainty, preds
    ):
        # Put objectives together
        objective = (
            self.uncertainty_weight * total_uncertainty
            + self.aleatoric_weight * aleatoric_uncertainty
            + self.epistemic_weight * epistemic_uncertainty
        )

        if self.VAE is not None and self.cond_mask is None and self.prior_weight > 0:
            try:
                prior_loglike = self.VAE.prior.log_prob(self.z).sum(dim=1)
            except:  # This mode is just for CondCLUE but the objective method is inherited
                prior_loglike = (
                    self.VAEAC.get_prior(self.original_x, self.cond_mask, flatten=False)
                    .log_prob(self.z)
                    .sum(dim=1)
                )
            objective += self.prior_weight * prior_loglike

        if self.latent_L2_weight != 0 and self.latent_L2_weight is not None:
            latent_dist = (
                F.mse_loss(self.z, self.z_init, reduction="none")
                .view(x.shape[0], -1)
                .sum(dim=1)
            )
            objective += self.latent_L2_weight * latent_dist

        if self.desired_preds is not None:
            pred_dist = self.pred_dist(preds).view(preds.shape[0], -1).sum(dim=1)
            objective += self.prediction_similarity_weight * pred_dist

        if self.distance_metric is not None:
            dist = (
                self.distance_metric(x, self.original_x).view(x.shape[0], -1).sum(dim=1)
            )
            objective += self.distance_weight * dist

            return objective, self.distance_weight * dist
        else:
            return objective, 0

    def optimise(self, min_steps=3, max_steps=25, n_early_stop=3):
        # Vectors to capture changes for this minibatch
        z_vec = [self.z.data.cpu().numpy()]
        x_vec = []
        uncertainty_vec = np.zeros((max_steps, self.z.shape[0]))
        aleatoric_vec = np.zeros((max_steps, self.z.shape[0]))
        epistemic_vec = np.zeros((max_steps, self.z.shape[0]))
        dist_vec = np.zeros((max_steps, self.z.shape[0]))
        cost_vec = np.zeros(
            (max_steps, self.z.shape[0])
        )  # this one doesnt consider the prior

        it_mask = np.zeros(self.z.shape[0])

        for step_idx in range(max_steps):
            self.optimizer.zero_grad()
            (
                total_uncertainty,
                aleatoric_uncertainty,
                epistemic_uncertainty,
                x,
                preds,
            ) = self.uncertainty_from_z()
            objective, w_dist = self.get_objective(
                x,
                total_uncertainty,
                aleatoric_uncertainty,
                epistemic_uncertainty,
                preds,
            )
            # We sum over features and over batch size in order to make dz invariant of batch (used to average over batch size)
            objective.sum(dim=0).backward()  # backpropagate

            self.optimizer.step()

            # save vectors
            uncertainty_vec[step_idx, :] = total_uncertainty.data.cpu().numpy()
            aleatoric_vec[step_idx, :] = aleatoric_uncertainty.data.cpu().numpy()
            epistemic_vec[step_idx, :] = epistemic_uncertainty.data.cpu().numpy()
            dist_vec[step_idx, :] = w_dist.data.cpu().numpy()
            cost_vec[step_idx, :] = objective.data.cpu().numpy()
            x_vec.append(
                x.data
            )  # we dont convert to numpy yet because we need x0 for L1
            z_vec.append(
                self.z.data.cpu().numpy()
            )  # this one is after gradient update while x is before

            it_mask = CLUE.update_stopvec(
                cost_vec, it_mask, step_idx, n_early_stop, min_steps
            )

        #  Generate final (or resulting s sample)

        x = self.VAE.regenerate(self.z, grad=False).data
        x_vec.append(x)
        x_vec = [i.cpu().numpy() for i in x_vec]  # convert x to numpy
        x_vec = np.stack(x_vec)
        z_vec = np.stack(z_vec)

        # Recover correct indexes using mask
        (
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            dist_vec,
            cost_vec,
            z_vec,
            x_vec,
        ) = CLUE.apply_stopvec(
            it_mask,
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            dist_vec,
            cost_vec,
            z_vec,
            x_vec,
            n_early_stop,
        )
        return (
            z_vec,
            x_vec,
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            cost_vec,
            dist_vec,
        )

    @staticmethod
    def update_stopvec(cost_vec, it_mask, step_idx, n_early_stop, min_steps):
        # TODO: can go in a CLUE parent class
        asymptotic_rel = (
            np.abs(cost_vec[step_idx - n_early_stop, :] - cost_vec[step_idx, :])
            < cost_vec[0, :] * 1e-2
        )
        asymptotic_abs = (
            np.abs(cost_vec[step_idx - n_early_stop, :] - cost_vec[step_idx, :]) < 1e-3
        )

        if step_idx > min_steps:
            condition_sum = asymptotic_rel + asymptotic_abs
        else:
            condition_sum = np.array([0])

        stop_vec = condition_sum.clip(max=1, min=0)

        to_mask = (it_mask == 0).astype(int) * stop_vec
        it_mask[to_mask == 1] = step_idx

        if (it_mask == 0).sum() == 0 and n_early_stop > 0:
            log.debug("iteration %d, all conditions met, stopping" % step_idx)
        return it_mask

    @staticmethod
    def apply_stopvec(
        it_mask,
        uncertainty_vec,
        epistemic_vec,
        aleatoric_vec,
        dist_vec,
        cost_vec,
        z_vec,
        x_vec,
        n_early_stop,
    ):
        # uncertainty_vec[step_idx, batch_size]
        it_mask = (it_mask - n_early_stop + 1).astype(int)
        for i in range(uncertainty_vec.shape[1]):
            if it_mask[i] > 0 and n_early_stop > 0:
                uncertainty_vec[it_mask[i] :, i] = uncertainty_vec[it_mask[i], i]
                epistemic_vec[it_mask[i] :, i] = epistemic_vec[it_mask[i], i]
                aleatoric_vec[it_mask[i] :, i] = aleatoric_vec[it_mask[i], i]
                cost_vec[it_mask[i] :, i] = cost_vec[it_mask[i], i]
                dist_vec[it_mask[i] :, i] = dist_vec[it_mask[i], i]
                z_vec[it_mask[i] :, i] = z_vec[it_mask[i], i]
                x_vec[it_mask[i] :, i] = x_vec[it_mask[i], i]
        return (
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            dist_vec,
            cost_vec,
            z_vec,
            x_vec,
        )

    def sample_explanations(
        self, n_explanations, init_std=0.15, min_steps=3, max_steps=25, n_early_stop=3
    ):
        # This creates a new first axis and stacks outputs there
        full_x_vec = []
        full_z_vec = []
        full_uncertainty_vec = []
        full_aleatoric_vec = []
        full_epistemic_vec = []
        full_dist_vec = []
        full_cost_vec = []

        for i in range(n_explanations):
            self.randomise_z_init(std=init_std)

            torch.autograd.set_detect_anomaly(False)

            (
                z_vec,
                x_vec,
                uncertainty_vec,
                epistemic_vec,
                aleatoric_vec,
                cost_vec,
                dist_vec,
            ) = self.optimise(
                min_steps=min_steps, max_steps=max_steps, n_early_stop=n_early_stop
            )

            full_x_vec.append(x_vec)
            full_z_vec.append(z_vec)
            full_uncertainty_vec.append(uncertainty_vec)
            full_aleatoric_vec.append(aleatoric_vec)
            full_epistemic_vec.append(epistemic_vec)
            full_dist_vec.append(dist_vec)
            full_cost_vec.append(cost_vec)

        full_x_vec = np.concatenate(np.expand_dims(full_x_vec, axis=0), axis=0)
        full_z_vec = np.concatenate(np.expand_dims(full_z_vec, axis=0), axis=0)
        full_cost_vec = np.concatenate(np.expand_dims(full_cost_vec, axis=0), axis=0)
        full_dist_vec = np.concatenate(np.expand_dims(full_dist_vec, axis=0), axis=0)
        full_uncertainty_vec = np.concatenate(
            np.expand_dims(full_uncertainty_vec, axis=0), axis=0
        )
        full_aleatoric_vec = np.concatenate(
            np.expand_dims(full_aleatoric_vec, axis=0), axis=0
        )
        full_epistemic_vec = np.concatenate(
            np.expand_dims(full_epistemic_vec, axis=0), axis=0
        )

        return (
            full_x_vec,
            full_z_vec,
            full_uncertainty_vec,
            full_aleatoric_vec,
            full_epistemic_vec,
            full_dist_vec,
            full_cost_vec,
        )

    @classmethod
    def batch_optimise(
        cls,
        VAE,
        BNN,
        original_x,
        uncertainty_weight,
        aleatoric_weight,
        epistemic_weight,
        prior_weight,
        distance_weight,
        latent_L2_weight,
        prediction_similarity_weight,
        lr,
        min_steps=3,
        max_steps=25,
        n_early_stop=3,
        batch_size=256,
        cond_mask=None,
        desired_preds=None,
        distance_metric=None,
        z_init=None,
        norm_MNIST=False,
        flatten_BNN=False,
        regression=False,
        prob_BNN=True,
        cuda=False,
    ):
        # This stacks outputs along the first (batch_size) axis
        full_x_vec = []
        full_z_vec = []
        full_uncertainty_vec = []
        full_aleatoric_vec = []
        full_epistemic_vec = []
        full_dist_vec = []
        full_cost_vec = []

        idx_iterator = generate_ind_batch(
            original_x.shape[0], batch_size=batch_size, random=False, roundup=True
        )
        for train_idx in idx_iterator:

            if z_init is not None:
                z_init_use = z_init[train_idx]
            else:
                z_init_use = z_init

            if desired_preds is not None:
                desired_preds_use = desired_preds[train_idx].data
            else:
                desired_preds_use = desired_preds

            CLUE_runner = cls(
                VAE,
                BNN,
                original_x[train_idx],
                uncertainty_weight,
                aleatoric_weight,
                epistemic_weight,
                prior_weight,
                distance_weight,
                latent_L2_weight,
                prediction_similarity_weight,
                lr,
                cond_mask=cond_mask,
                distance_metric=distance_metric,
                z_init=z_init_use,
                norm_MNIST=norm_MNIST,
                desired_preds=desired_preds_use,
                flatten_BNN=flatten_BNN,
                regression=regression,
                prob_BNN=prob_BNN,
                cuda=cuda,
            )

            (
                z_vec,
                x_vec,
                uncertainty_vec,
                epistemic_vec,
                aleatoric_vec,
                cost_vec,
                dist_vec,
            ) = CLUE_runner.optimise(
                min_steps=min_steps, max_steps=max_steps, n_early_stop=n_early_stop
            )

            full_x_vec.append(x_vec)
            full_z_vec.append(z_vec)
            full_uncertainty_vec.append(uncertainty_vec)
            full_aleatoric_vec.append(aleatoric_vec)
            full_epistemic_vec.append(epistemic_vec)
            full_dist_vec.append(dist_vec)
            full_cost_vec.append(cost_vec)

        full_x_vec = np.concatenate(full_x_vec, axis=1)
        full_z_vec = np.concatenate(full_z_vec, axis=1)
        full_cost_vec = np.concatenate(full_cost_vec, axis=1)
        full_dist_vec = np.concatenate(full_dist_vec, axis=1)
        full_uncertainty_vec = np.concatenate(full_uncertainty_vec, axis=1)
        full_aleatoric_vec = np.concatenate(full_aleatoric_vec, axis=1)
        full_epistemic_vec = np.concatenate(full_epistemic_vec, axis=1)

        return (
            full_x_vec,
            full_z_vec,
            full_uncertainty_vec,
            full_aleatoric_vec,
            full_epistemic_vec,
            full_dist_vec,
            full_cost_vec,
        )


class conditional_CLUE(CLUE):
    def __init__(
        self,
        VAEAC,
        BNN,
        original_x,
        uncertainty_weight,
        aleatoric_weight,
        epistemic_weight,
        prior_weight,
        distance_weight,
        lr,
        cond_mask=None,
        distance_metric=None,
        z_init=None,
        norm_MNIST=False,
        flatten_BNN=False,
        regression=False,
        cuda=False,
    ):

        super(conditional_CLUE, self).__init__(
            VAEAC,
            BNN,
            original_x,
            uncertainty_weight,
            aleatoric_weight,
            epistemic_weight,
            prior_weight,
            distance_weight,
            lr,
            cond_mask,
            distance_metric,
            z_init,
            norm_MNIST,
            flatten_BNN,
            regression,
            cuda,
        )
        self.cond_mask = cond_mask.type(original_x.type())
        self.VAEAC = VAEAC
        self.prior_weight = 0

    def uncertainty_from_z(self):

        x = self.VAEAC.regenerate(self.z, grad=True)
        x = x * self.cond_mask + self.original_x * (1 - self.cond_mask)

        if self.flatten_BNN:
            to_BNN = x.view(x.shape[0], -1)
        else:
            to_BNN = x

        if self.norm_MNIST:
            to_BNN = MNIST_mean_std_norm(to_BNN)

        if self.regression:
            mu_vec, std_vec = self.BNN.sample_predict(to_BNN, Nsamples=0, grad=True)
            (
                total_uncertainty,
                aleatoric_uncertainty,
                epistemic_uncertainty,
            ) = decompose_std_gauss(mu_vec, std_vec)
        else:
            probs = self.BNN.sample_predict(to_BNN, Nsamples=0, grad=False)
            (
                total_uncertainty,
                aleatoric_uncertainty,
                epistemic_uncertainty,
            ) = decompose_entropy_cat(probs)

        return total_uncertainty, aleatoric_uncertainty, epistemic_uncertainty, x

    def optimise(self, min_steps=3, max_steps=25, n_early_stop=3):
        # Vectors to capture changes for this minibatch
        z_vec = [self.z.data.cpu().numpy()]
        x_vec = []
        uncertainty_vec = np.zeros((max_steps, self.z.shape[0]))
        aleatoric_vec = np.zeros((max_steps, self.z.shape[0]))
        epistemic_vec = np.zeros((max_steps, self.z.shape[0]))
        dist_vec = np.zeros((max_steps, self.z.shape[0]))
        cost_vec = np.zeros(
            (max_steps, self.z.shape[0])
        )  # this one doesnt consider the prior

        it_mask = np.zeros(self.z.shape[0])

        for step_idx in range(max_steps):
            self.optimizer.zero_grad()
            (
                total_uncertainty,
                aleatoric_uncertainty,
                epistemic_uncertainty,
                x,
            ) = self.uncertainty_from_z()
            objective, w_dist = self.get_objective(
                x, total_uncertainty, aleatoric_uncertainty, epistemic_uncertainty
            )
            objective.mean(dim=0).backward()  # backpropagate

            self.optimizer.step()

            # save vectors
            uncertainty_vec[step_idx, :] = total_uncertainty.data.cpu().numpy()
            aleatoric_vec[step_idx, :] = aleatoric_uncertainty.data.cpu().numpy()
            epistemic_vec[step_idx, :] = epistemic_uncertainty.data.cpu().numpy()
            dist_vec[step_idx, :] = w_dist.data.cpu().numpy()
            cost_vec[step_idx, :] = objective.data.cpu().numpy()
            x_vec.append(
                x.data
            )  # we dont convert to numpy yet because we need x0 for L1
            z_vec.append(
                self.z.data.cpu().numpy()
            )  # this one is after gradient update while x is before

            it_mask = CLUE.update_stopvec(
                cost_vec, it_mask, step_idx, n_early_stop, min_steps
            )

        #  Generate final (or resulting s sample)

        x = self.VAE.regenerate(self.z, grad=False).data
        x = x * self.cond_mask + self.original_x * (1 - self.cond_mask)
        x_vec.append(x)
        x_vec = [i.cpu().numpy() for i in x_vec]  # convert x to numpy
        x_vec = np.stack(x_vec)
        z_vec = np.stack(z_vec)

        # Recover correct indexes using mask
        (
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            dist_vec,
            cost_vec,
            z_vec,
            x_vec,
        ) = CLUE.apply_stopvec(
            it_mask,
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            dist_vec,
            cost_vec,
            z_vec,
            x_vec,
            n_early_stop,
        )
        return (
            z_vec,
            x_vec,
            uncertainty_vec,
            epistemic_vec,
            aleatoric_vec,
            cost_vec,
            dist_vec,
        )



from typing import Any, Dict

import pandas as pd

from carla.models.api import MLModel
from carla.recourse_methods.api import RecourseMethod
from carla.recourse_methods.catalog.face.library import graph_search
from carla.recourse_methods.processing import (
    check_counterfactuals,
    encode_feature_names,
    merge_default_parameters,
)


class Face(RecourseMethod):
    """
    Implementation of FACE from Poyiadzi et.al. [1]_.

    Parameters
    ----------
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Dictionary containing hyperparameters. See notes below for its contents.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Notes
    -----
    - Hyperparams
        Hyperparameter contains important information for the recourse method to initialize.
        Please make sure to pass all values as dict with the following keys.

        * "mode": {"knn", "epsilon"},
            Decides over type of FACE
        * "fraction": float [0 < x < 1]
            determines fraction of data set to be used to construct neighbourhood graph

    .. [1] Rafael Poyiadzi, Kacper Sokol, Raul Santos-Rodriguez, Tijl De Bie, and Peter Flach. 2020. In Proceedings
            of the AAAI/ACM Conference on AI, Ethics, and Society (AIES)
    """

    _DEFAULT_HYPERPARAMS = {"mode": None, "fraction": 0.1}

    def __init__(self, mlmodel: MLModel, hyperparams: Dict[str, Any]) -> None:
        super().__init__(mlmodel)

        checked_hyperparams = merge_default_parameters(
            hyperparams, self._DEFAULT_HYPERPARAMS
        )
        self.mode = checked_hyperparams["mode"]
        self.fraction = checked_hyperparams["fraction"]

        # Normalize and encode data
        self._df_enc_norm = self.encode_normalize_order_factuals(self._mlmodel.data.raw)

        self._immutables = encode_feature_names(
            self._mlmodel.data.immutables, self._mlmodel.feature_input_order
        )

    @property
    def fraction(self) -> float:
        """
        Controls the fraction of the used dataset to build graph on.

        Returns
        -------
        float
        """
        return self._fraction

    @fraction.setter
    def fraction(self, x: float) -> None:
        if 0 < x < 1:
            self._fraction = x
        else:
            raise ValueError("Fraction has to be between 0 and 1")

    @property
    def mode(self) -> str:
        """
        Sets and changes the type of FACE. {"knn", "epsilon"}

        Returns
        -------
        str
        """
        return self._mode

    @mode.setter
    def mode(self, mode: str) -> None:
        if mode in ["knn", "epsilon"]:
            self._mode = mode
        else:
            raise ValueError("Mode has to be either knn or epsilon")

    def get_counterfactuals(self, factuals: pd.DataFrame) -> pd.DataFrame:
        # Normalize and encode factual
        df_enc_norm_fact = self.encode_normalize_order_factuals(factuals)

        # >drop< factuals from dataset to prevent duplicates,
        # >reorder< and >add< factuals to top; necessary in order to use the index
        df_enc_norm_data = self._df_enc_norm.copy()
        cond = df_enc_norm_data.isin(df_enc_norm_fact).values
        df_enc_norm_data = df_enc_norm_data.drop(df_enc_norm_data[cond].index)
        df_enc_norm_data = pd.concat(
            [df_enc_norm_fact, df_enc_norm_data], ignore_index=True
        )

        list_cfs = []
        for i in range(df_enc_norm_fact.shape[0]):
            cf = graph_search(
                df_enc_norm_data,
                i,
                self._immutables,
                self._mlmodel,
                mode=self._mode,
                frac=self._fraction,
            )
            list_cfs.append(cf)

        df_cfs = check_counterfactuals(self._mlmodel, list_cfs)

        return df_cfs



import pandas as pd

from carla.models.api import MLModel
from carla.recourse_methods.api import RecourseMethod
from carla.recourse_methods.catalog.growing_spheres.library import (
    growing_spheres_search,
)
from carla.recourse_methods.processing import (
    check_counterfactuals,
    encode_feature_names,
)


class GrowingSpheres(RecourseMethod):
    """
    Implementation of Growing Spheres from Laugel et.al. [1]_.

    Parameters
    ----------
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Growing Spheeres needs no hyperparams.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Notes
    -----
    - Restrictions
        Growing Spheres works at the moment only for data with dropped first column of binary categorical features.

    .. [1] Thibault Laugel, Marie-Jeanne Lesot, Christophe Marsala, Xavier Renard, and Marcin Detyniecki. 2017.
            Inverse Classification for Comparison-based Interpretability in Machine Learning.
            arXiv preprint arXiv:1712.08443(2017).
    """

    def __init__(self, mlmodel: MLModel, hyperparams=None) -> None:
        super().__init__(mlmodel)

        self._immutables = encode_feature_names(
            self._mlmodel.data.immutables, self._mlmodel.feature_input_order
        )
        self._mutables = [
            feature
            for feature in self._mlmodel.feature_input_order
            if feature not in self._immutables
        ]
        self._continuous = self._mlmodel.data.continous
        self._categoricals_enc = encode_feature_names(
            self._mlmodel.data.categoricals, self._mlmodel.feature_input_order
        )

    def get_counterfactuals(self, factuals: pd.DataFrame) -> pd.DataFrame:
        # Normalize and encode data
        df_enc_norm_fact = self.encode_normalize_order_factuals(factuals)

        list_cfs = []
        for index, row in df_enc_norm_fact.iterrows():
            counterfactual = growing_spheres_search(
                row,
                self._mutables,
                self._immutables,
                self._continuous,
                self._categoricals_enc,
                self._mlmodel.feature_input_order,
                self._mlmodel,
            )
            list_cfs.append(counterfactual)

        df_cfs = check_counterfactuals(self._mlmodel, list_cfs)

        return df_cfs



# flake8: noqa
from __future__ import division

import numpy as np
import torch
from torch.distributions import Categorical, constraints
from torch.distributions.normal import Normal
from torch.distributions.transformed_distribution import TransformedDistribution
from torch.distributions.transforms import StickBreakingTransform
from torch.nn import Module
from torch.nn.functional import softmax, softplus


# TODO: this might be better changed to a softmax transform instead of stick breaking but would need to look into it
class LogisticNormal(TransformedDistribution):
    r"""
    Creates a logistic-normal distribution parameterized by :attr:`loc` and :attr:`scale`
    that define the base `Normal` distribution transformed with the
    `StickBreakingTransform` such that::
        X ~ LogisticNormal(loc, scale)
        Y = log(X / (1 - X.cumsum(-1)))[..., :-1] ~ Normal(loc, scale)
    Args:
        loc (float or Tensor): mean of the base distribution
        scale (float or Tensor): standard deviation of the base distribution
    Example::
        >>> # logistic-normal distributed with mean=(0, 0, 0) and stddev=(1, 1, 1)
        >>> # of the base Normal distribution
        >>> m = distributions.LogisticNormal(torch.tensor([0.0] * 3), torch.tensor([1.0] * 3))
        >>> m.sample()
        tensor([ 0.7653,  0.0341,  0.0579,  0.1427])
    """
    arg_constraints = {"loc": constraints.real, "scale": constraints.positive}
    support = constraints.simplex
    has_rsample = True

    def __init__(self, loc, scale, validate_args=None):
        base_dist = Normal(loc, scale)
        super(LogisticNormal, self).__init__(
            base_dist, StickBreakingTransform(), validate_args=validate_args
        )
        # Adjust event shape since StickBreakingTransform adds 1 dimension
        self._event_shape = torch.Size([s + 1 for s in self._event_shape])

    def expand(self, batch_shape, _instance=None):
        new = self._get_checked_instance(LogisticNormal, _instance)
        return super(LogisticNormal, self).expand(batch_shape, _instance=new)

    @property
    def loc(self):
        return self.base_dist.loc

    @property
    def scale(self):
        return self.base_dist.scale


def normal_parse_params(params, min_sigma=1e-3):
    """
    Take a Tensor (e. g. neural network output) and return
    torch.distributions.Normal distribution.
    This Normal distribution is component-wise independent,
    and its dimensionality depends on the input shape.
    First half of channels is mean of the distribution,
    the softplus of the second half is std (sigma), so there is
    no restrictions on the input tensor.
    min_sigma is the minimal value of sigma. I. e. if the above
    softplus is less than min_sigma, then sigma is clipped
    from below with value min_sigma. This regularization
    is required for the numerical stability and may be considered
    as a neural network architecture choice without any change
    to the probabilistic model.
    """
    n = params.shape[0]
    d = params.shape[1]
    mu = params[:, : d // 2]
    sigma_params = params[:, d // 2 :]
    sigma = softplus(sigma_params)
    sigma = sigma.clamp(min=min_sigma)
    distr = Normal(mu, sigma)
    return distr


def categorical_parse_params_column(params, min_prob=1e-2):
    """
    Take a Tensor (e. g. a part of neural network output) and return
    torch.distributions.Categorical distribution.
    The input tensor after applying softmax over the last axis contains
    a batch of the categorical probabilities. So there are no restrictions
    on the input tensor.
    Technically, this function treats the last axis as the categorical
    probabilities, but Categorical takes only 2D input where
    the first axis is the batch axis and the second one corresponds
    to the probabilities, so practically the function requires 2D input
    with the batch of probabilities for one categorical feature.
    min_prob is the minimal probability for each class.
    After clipping the probabilities from below they are renormalized
    in order to be a valid distribution. This regularization
    is required for the numerical stability and may be considered
    as a neural network architecture choice without any change
    to the probabilistic model.
    """
    params = softmax(params, -1)
    params = params.clamp(min_prob)
    params = params / params.sum(-1, keepdim=True)
    distr = Categorical(probs=params)
    return distr


class GaussianLoglike(Module):
    """
    Compute reconstruction log probability of groundtruth given
    a tensor of Gaussian distribution parameters and a mask.
    Gaussian distribution parameters are output of a neural network
    without any restrictions, the minimal sigma value is clipped
    from below to min_sigma (default: 1e-2) in order not to overfit
    network on some exact pixels.
    The first half of channels corresponds to mean, the second half
    corresponds to std. See normal_parse_parameters for more info.
    This layer doesn't work with NaNs in the data, it is used for
    inpainting. Roughly speaking, this loss is similar to L2 loss.
    Returns a vector of log probabilities for each object of the batch.
    """

    def __init__(self, min_sigma=1e-2):
        super(GaussianLoglike, self).__init__()
        self.min_sigma = min_sigma

    def forward(self, distr_params, groundtruth, mask=None):
        distr = normal_parse_params(distr_params, self.min_sigma)
        if mask is not None:
            log_probs = distr.log_prob(groundtruth) * mask
        else:
            log_probs = distr.log_prob(groundtruth)
        return log_probs.view(groundtruth.shape[0], -1).sum(-1)


# functions for BNN with gauss output:
def diagonal_gauss_loglike(x, mu, sigma):
    # note that we can just treat each dim as isotropic and then do sum
    cte_term = -(0.5) * np.log(2 * np.pi)
    det_sig_term = -torch.log(sigma)
    inner = (x - mu) / sigma
    dist_term = -(0.5) * (inner ** 2)
    log_px = (cte_term + det_sig_term + dist_term).sum(dim=1, keepdim=False)
    return log_px


def get_rms(mu, y, y_means, y_stds):
    x_un = mu * y_stds + y_means
    y_un = y * y_stds + y_means
    return torch.sqrt(((x_un - y_un) ** 2).sum() / y.shape[0])


def gmm_likelihood(x, mu_vec, sigma_vec):
    weight_factor = np.log(mu_vec.shape[0])
    loglike_terms = []
    for i in range(mu_vec.shape[0]):
        loglike_terms.append(diagonal_gauss_loglike(x, mu_vec[i], sigma_vec[i]))
    loglike_terms = torch.cat(loglike_terms, dim=0)

    out = torch.logsumexp(loglike_terms, dim=0) - weight_factor
    return out


# TODO: rename this function to something gaussian
def get_loglike(mu, sigma, y, y_means, y_stds, gmm=False):
    mu_un = mu * y_stds + y_means
    y_un = y * y_stds + y_means
    sigma_un = sigma * y_stds
    if gmm:
        ll = gmm_likelihood(y_un, mu_un, sigma_un)
    else:
        ll = diagonal_gauss_loglike(y_un, mu_un, sigma_un)
    return ll.mean(dim=0)  # mean over datapoints


def complete_logit_norm_vec(vec):
    last_term = 1 - vec.sum(dim=1)
    cvec = torch.cat((vec, last_term), dim=1)
    return cvec


def marginal_std(mu, sigma):  # This is for outputs from NN and GMM var estimation
    """Obtain the std of a GMM with isotropic components"""
    # probs (Nsamples, batch_size, classes)
    marg_var = (sigma ** 2).mean(dim=0) + ((mu ** 2).mean(dim=0) - mu.mean(dim=0) ** 2)
    return torch.sqrt(marg_var)


total_std = marginal_std


def decompose_var_gauss(mu, sigma, sum_dims=True):
    # probs (Nsamples, batch_size, output_sims)
    aleatoric_var = (sigma ** 2).mean(dim=0)
    epistemic_var = (mu ** 2).mean(dim=0) - mu.mean(dim=0) ** 2
    total_var = aleatoric_var + epistemic_var
    if sum_dims:
        aleatoric_var = aleatoric_var.sum(dim=1)
        epistemic_var = epistemic_var.sum(dim=1)
        total_var = total_var.sum(dim=1)
    return total_var, aleatoric_var, epistemic_var


def decompose_std_gauss(mu, sigma, sum_dims=True):
    # probs (Nsamples, batch_size, output_sims)
    aleatoric_var = (sigma ** 2).mean(dim=0)
    epistemic_var = (mu ** 2).mean(dim=0) - mu.mean(dim=0) ** 2
    total_var = aleatoric_var + epistemic_var
    if sum_dims:
        aleatoric_var = aleatoric_var.sum(dim=1)
        epistemic_var = epistemic_var.sum(dim=1)
        total_var = total_var.sum(dim=1)
    return total_var.sqrt(), aleatoric_var.sqrt(), epistemic_var.sqrt()


def decompose_entropy_cat(probs, eps=1e-10):
    # probs (Nsamples, batch_size, classes)
    posterior_preds = probs.mean(dim=0, keepdim=False)
    total_entropy = -(posterior_preds * torch.log(posterior_preds + eps)).sum(
        dim=1, keepdim=False
    )

    sample_preds_entropy = -(probs * torch.log(probs + eps)).sum(dim=2, keepdim=False)
    aleatoric_entropy = sample_preds_entropy.mean(dim=0, keepdim=False)

    epistemic_entropy = total_entropy - aleatoric_entropy

    # returns (batch_size)
    return total_entropy, aleatoric_entropy, epistemic_entropy



from typing import Any, Dict, List

import pandas as pd

from carla.data.load_catalog import load_catalog

from ..api import Data
from .load_data import load_dataset


class DataCatalog(Data):
    """
    Use already implemented datasets.

    Parameters
    ----------
    data_name : {'adult', 'compas', 'give_me_some_credit'}
        Used to get the correct dataset from online repository.

    Returns
    -------
    None
    """

    def __init__(self, data_name: str):
        self.name = data_name

        catalog_content = ["continous", "categorical", "immutable", "target"]
        self.catalog: Dict[str, Any] = load_catalog(  # type: ignore
            "data_catalog.yaml", data_name, catalog_content
        )

        for key in ["continous", "categorical", "immutable"]:
            if self.catalog[key] is None:
                self.catalog[key] = []

        self._raw: pd.DataFrame = load_dataset(data_name)

    @property
    def categoricals(self) -> List[str]:
        return self.catalog["categorical"]

    @property
    def continous(self) -> List[str]:
        return self.catalog["continous"]

    @property
    def immutables(self) -> List[str]:
        return self.catalog["immutable"]

    @property
    def target(self) -> str:
        return self.catalog["target"]

    @property
    def raw(self) -> pd.DataFrame:
        return self._raw.copy()



from __future__ import division

import numpy as np
import torch.backends.cudnn as cudnn
from torch.distributions import kl_divergence
from torch.distributions.normal import Normal

from carla import log
from carla.recourse_methods.catalog.clue.library.clue_ml.src.gauss_cat import *
from carla.recourse_methods.catalog.clue.library.clue_ml.src.probability import (
    normal_parse_params,
)
from carla.recourse_methods.catalog.clue.library.clue_ml.src.radam import RAdam
from carla.recourse_methods.catalog.clue.library.clue_ml.src.utils import (
    BaseNet,
    to_variable,
)

from .models import MLP_preact_generator_net, MLP_preact_recognition_net

# TODO: implement for std changeable gaussian instead of rms


class VAE_gauss_cat(nn.Module):
    def __init__(self, input_dim_vec, width, depth, latent_dim, pred_sig=False):
        super(VAE_gauss_cat, self).__init__()

        input_dim = 0
        self.input_dim_vec = input_dim_vec
        for e in input_dim_vec:
            input_dim += e

        self.encoder = MLP_preact_recognition_net(input_dim, width, depth, latent_dim)
        if pred_sig:
            raise NotImplementedError()
        else:
            self.decoder = MLP_preact_generator_net(input_dim, width, depth, latent_dim)
            self.rec_loglike = rms_cat_loglike(self.input_dim_vec, reduction="none")
        self.pred_sig = pred_sig

    def encode(self, x):
        """Works with flattened representATION"""
        approx_post_params = self.encoder(x)
        approx_post = normal_parse_params(approx_post_params, 1e-3)
        return approx_post

    def decode(self, z_sample):
        """Works with flattened representATION"""
        rec_params = self.decoder(z_sample)
        return rec_params

    def vlb(self, prior, approx_post, x, rec_params):
        """Works with flattened representATION"""
        if self.pred_sig:
            pass
        else:
            rec = self.rec_loglike(rec_params, x).view(x.shape[0], -1).sum(-1)
        kl = kl_divergence(approx_post, prior).view(x.shape[0], -1).sum(-1)
        return rec - kl

    def iwlb(self, prior, approx_post, x, K=50):
        estimates = []
        for i in range(K):
            latent = approx_post.rsample()
            rec_params = self.decode(latent)
            if self.pred_sig:
                pass
            else:
                rec_loglike = (
                    self.rec_loglike(rec_params, x).view(x.shape[0], -1).sum(-1)
                )

            prior_log_prob = prior.log_prob(latent)
            prior_log_prob = prior_log_prob.view(x.shape[0], -1)
            prior_log_prob = prior_log_prob.sum(-1)

            proposal_log_prob = approx_post.log_prob(latent)
            proposal_log_prob = proposal_log_prob.view(x.shape[0], -1)
            proposal_log_prob = proposal_log_prob.sum(-1)

            estimate = rec_loglike + prior_log_prob - proposal_log_prob
            estimates.append(estimate[:, None])

        return torch.logsumexp(torch.cat(estimates, 1), 1) - np.log(K)


class VAE_gauss_cat_net(BaseNet):
    def __init__(
        self,
        input_dim_vec,
        width,
        depth,
        latent_dim,
        pred_sig=False,
        lr=1e-3,
        cuda=True,
        flatten=True,
    ):
        super(VAE_gauss_cat_net, self).__init__()
        log.info("VAE_gauss_net")

        self.cuda = cuda
        self.input_dim = 0
        self.input_dim_vec = input_dim_vec
        for e in self.input_dim_vec:
            self.input_dim += e
        self.flatten = flatten
        if not self.flatten:
            pass

        self.width = width
        self.depth = depth
        self.latent_dim = latent_dim
        self.lr = lr
        self.pred_sig = pred_sig

        self.create_net()
        self.create_opt()
        self.epoch = 0
        self.schedule = None

        if self.cuda:
            self.prior = self.prior = Normal(
                loc=torch.zeros(latent_dim).cuda(), scale=torch.ones(latent_dim).cuda()
            )
        else:
            self.prior = Normal(
                loc=torch.zeros(latent_dim), scale=torch.ones(latent_dim)
            )
        self.vlb_scale = 1 / len(
            self.input_dim_vec
        )  # scale for dimensions of input so we can use same LR always

    def create_net(self):
        torch.manual_seed(42)
        torch.cuda.manual_seed(42)
        self.model = VAE_gauss_cat(
            self.input_dim_vec, self.width, self.depth, self.latent_dim, self.pred_sig
        )
        if self.cuda:
            self.model = self.model.cuda()
            cudnn.benchmark = True
        log.info("Total params: %.2fM" % (self.get_nb_parameters() / 1000000.0))

    def create_opt(self):
        self.optimizer = RAdam(self.model.parameters(), lr=self.lr)

    def fit(self, x):
        self.set_mode_train(train=True)

        if self.flatten:
            x_flat = gauss_cat_to_flat(x, self.input_dim_vec)
        else:
            x_flat = x
            x = flat_to_gauss_cat(x, self.input_dim_vec)

        x, x_flat = to_variable(var=(x, x_flat), cuda=self.cuda)
        self.optimizer.zero_grad()

        approx_post = self.model.encode(x_flat)
        z_sample = approx_post.rsample()
        rec_params = self.model.decode(z_sample)

        vlb = self.model.vlb(self.prior, approx_post, x, rec_params)
        loss = (-vlb * self.vlb_scale).mean()

        loss.backward()
        self.optimizer.step()

        return vlb.mean().item(), rec_params

    def eval(self, x, sample=False):
        self.set_mode_train(train=False)

        if self.flatten:
            x_flat = gauss_cat_to_flat(x, self.input_dim_vec)
        else:
            x_flat = x
            x = flat_to_gauss_cat(x, self.input_dim_vec)

        x, x_flat = to_variable(var=(x, x_flat), cuda=self.cuda)
        approx_post = self.model.encode(x_flat)
        if sample:
            z_sample = approx_post.sample()
        else:
            z_sample = approx_post.loc
        rec_params = self.model.decode(z_sample)

        vlb = self.model.vlb(self.prior, approx_post, x, rec_params)

        return vlb.mean().item(), rec_params

    def eval_iw(self, x, k=50):
        self.set_mode_train(train=False)
        if self.flatten:
            x_flat = gauss_cat_to_flat(x, self.input_dim_vec)
        else:
            x_flat = x
            x = flat_to_gauss_cat(x, self.input_dim_vec)

        x, x_flat = to_variable(var=(x, x_flat), cuda=self.cuda)

        approx_post = self.model.recognition_encode(x)

        iw_lb = self.model.iwlb(self.prior, approx_post, x, k)
        return iw_lb.mean().item()

    def recongnition(self, x, grad=False, flatten=None):
        if flatten is None:
            flatten = self.flatten
        if flatten and grad:
            raise Exception("flatten and grad options are not compatible")
        self.set_mode_train(train=False)
        if flatten:
            x = gauss_cat_to_flat(x, self.input_dim_vec)
        if grad:
            if not x.requires_grad:
                x.requires_grad = True
        else:
            (x,) = to_variable(var=(x,), volatile=True, cuda=self.cuda)
        approx_post = self.model.encode(x)
        return approx_post

    def regenerate(self, z, grad=False, unflatten=False):
        if unflatten and grad:
            raise Exception("flatten and grad options are not compatible")
        self.set_mode_train(train=False)
        if grad:
            if not z.requires_grad:
                z.requires_grad = True
        else:
            (z,) = to_variable(var=(z,), volatile=True, cuda=self.cuda)
        out = self.model.decode(z)

        if unflatten:
            out = flat_to_gauss_cat(out, self.input_dim_vec)
        else:
            out = selective_softmax(out, self.input_dim_vec, grad=grad)

        if self.pred_sig:
            raise Exception("Not implemented")
        else:
            return out



from __future__ import division

from typing import Dict

import torch
import torch.nn as nn


def MLPBlock(width):
    return SkipConnection(
        nn.Linear(width, width), nn.ReLU(), nn.BatchNorm1d(num_features=width)
    )


def preact_MLPBlock(width):
    return SkipConnection(
        nn.ReLU(),
        nn.BatchNorm1d(num_features=width),
        nn.Linear(width, width),
    )


def leaky_MLPBlock(width):
    return SkipConnection(
        nn.Linear(width, width), nn.LeakyReLU(), nn.BatchNorm1d(num_features=width)
    )


def preact_leaky_MLPBlock(width):
    return SkipConnection(
        nn.LeakyReLU(),
        nn.BatchNorm1d(num_features=width),
        nn.Linear(width, width),
    )


class ResBlock(nn.Module):
    """
    Usual full pre-activation ResNet bottleneck block.
    For more information see
    He, K., Zhang, X., Ren, S., & Sun, J. (2016, October).
    Identity mappings in deep residual networks.
    European Conference on Computer Vision (pp. 630-645).
    Springer, Cham.
    ArXiv link: https://arxiv.org/abs/1603.05027
    """

    def __init__(self, outer_dim, inner_dim):
        super(ResBlock, self).__init__()
        self.net = nn.Sequential(
            nn.BatchNorm2d(outer_dim),
            nn.LeakyReLU(),
            nn.Conv2d(outer_dim, inner_dim, 1),
            nn.BatchNorm2d(inner_dim),
            nn.LeakyReLU(),
            nn.Conv2d(inner_dim, inner_dim, 3, 1, 1),
            nn.BatchNorm2d(inner_dim),
            nn.LeakyReLU(),
            nn.Conv2d(inner_dim, outer_dim, 1),
        )

    def forward(self, input):
        return input + self.net(input)


class SkipConnection(nn.Module):
    """
    Skip-connection over the sequence of layers in the constructor.
    The module passes input data sequentially through these layers
    and then adds original data to the result.
    """

    def __init__(self, *args):
        super(SkipConnection, self).__init__()
        self.inner_net = nn.Sequential(*args)

    def forward(self, input):
        return input + self.inner_net(input)


class MemoryLayer(nn.Module):
    """
    If output=False, this layer stores its input in a static class dictionary
    `storage` with the key `id` and then passes the input to the next layer.
    If output=True, this layer takes stored tensor from a static storage.
    If add=True, it returns sum of the stored vector and an input,
    otherwise it returns their concatenation.
    If the tensor with specified `id` is not in `storage` when the layer
    with output=True is called, it would cause an exception.
    The layer is used to make skip-connections inside nn.Sequential network
    or between several nn.Sequential networks without unnecessary code
    complication.
    The usage pattern is
    ```
        net1 = nn.Sequential(
            MemoryLayer('#1'),
            MemoryLayer('#0.1'),
            nn.Linear(512, 256),
            nn.LeakyReLU(),
            MemoryLayer('#0.1', output=True, add=False),
            # here add cannot be True because the dimensions mismatch
            nn.Linear(768, 256),
            # the dimension after the concatenation with skip-connection
            # is 512 + 256 = 768
        )
        net2 = nn.Sequential(
            nn.Linear(512, 512),
            MemoryLayer('#1', output=True, add=True),
            ...
        )
        b = net1(a)
        d = net2(c)
        # net2 must be called after net1,
        # otherwise tensor '#1' will not be in `storage`
    ```
    """

    storage: Dict = {}

    def __init__(self, id, output=False, add=False):
        super(MemoryLayer, self).__init__()
        self.id = id
        self.output = output
        self.add = add

    def forward(self, input):
        if not self.output:
            self.storage[self.id] = input
            return input
        else:
            if self.id not in self.storage:
                err = "MemoryLayer: id '%s' is not initialized. "
                err += "You must execute MemoryLayer with the same id "
                err += "and output=False before this layer."
                raise ValueError(err)
            stored = self.storage[self.id]
            if not self.add:
                data = torch.cat([input, stored], 1)
            else:
                data = input + stored
            return data



import os
from typing import Dict

import numpy as np
import pandas as pd
import torch
from sklearn.model_selection import train_test_split

from carla.data.api import Data
from carla.models.api import MLModel
from carla.recourse_methods.api import RecourseMethod
from carla.recourse_methods.catalog.clue.library import (
    VAE_gauss_cat_net,
    training,
    vae_gradient_search,
)
from carla.recourse_methods.processing import check_counterfactuals
from carla.recourse_methods.processing.counterfactuals import merge_default_parameters


class Clue(RecourseMethod):
    """
    Implementation of CLUE from Antorán et.al. [1]_.
    CLUE needs an variational autoencoder to generate counterfactual examples.
    By setting the train_ae key to True in hyperparams, a Pytorch VAE will be trained.

    Parameters
    ----------
    data : data.api.Data
            Underlying dataset we want to build counterfactuals for.
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Dictionary containing hyperparameters. See notes below for its contents.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Notes
    -----
    - Hyperparams
        Hyperparameter contains important information for the recourse method to initialize.
        Please make sure to pass all values as dict with the following keys.

        * "data_name": str
            Identifies the loaded or saved autoencoder model
        * "train_vae": bool
            Decides whether to load or train a vae
        * "width": int
            Structure for VAE
        * "depth": int
            Structure for VAE
        * "latent_dim": int
            Structure for VAE
        * "batch_size": int
            Structure for VAE
        * "epochs": int
            Structure for VAE
        * "lr": int
            Structure for VAE
        * "early_stop": int
            Structure for VAE

    .. [1] Javier Antorán, Umang Bhatt, Tameem Adel, Adrian Weller, and José Miguel Hernández-Lobato.
            Getting a CLUE: A Method for Explaining Uncertainty Estimates. In International Conference on
            Learning Representations (ICLR).
    """

    _DEFAULT_HYPERPARAMS = {
        "data_name": None,
        "train_vae": True,
        "width": 10,
        "depth": 3,
        "latent_dim": 12,
        "batch_size": 64,
        "epochs": 1,
        "lr": 0.001,
        "early_stop": 10,
    }

    def __init__(self, data: Data, mlmodel: MLModel, hyperparams: Dict) -> None:
        super().__init__(mlmodel)

        # get hyperparameter
        checked_hyperparams = merge_default_parameters(
            hyperparams, self._DEFAULT_HYPERPARAMS
        )
        self._vae_training = checked_hyperparams["train_vae"]
        self._width = checked_hyperparams["width"]
        self._depth = checked_hyperparams["depth"]
        self._latent_dim = checked_hyperparams["latent_dim"]
        self._data_name = checked_hyperparams["data_name"]
        self._batch_size = checked_hyperparams["batch_size"]
        self._epochs = checked_hyperparams["epochs"]
        self._lr = checked_hyperparams["lr"]
        self._early_stop = checked_hyperparams["early_stop"]
        self._continous = self._mlmodel.data.continous
        self._categorical = self._mlmodel.data.categoricals

        # get input dimension
        # indicate dimensions of inputs -- input_dim_vec: (if binary = 2; if continuous = 1)
        input_dims_continuous = list(np.repeat(1, len(self._mlmodel.data.continous)))
        input_dims_binary = list(np.repeat(2, len(self._mlmodel.data.categoricals)))
        self._input_dimension = input_dims_continuous + input_dims_binary

        # normalize and encode data
        self._df_norm_enc_data = self.encode_normalize_order_factuals(data.raw)

        # load autoencoder
        self._vae = self._load_vae()

    def _load_vae(self):
        # save_path
        path = os.environ.get(
            "CF_MODELS",
            os.path.join(
                "~",
                "carla",
                "models",
                "autoencoders",
                "clue",
                "fc_VAE_{}_models".format(self._data_name),
            ),
        )

        if not os.path.isfile(path) and not self._train_vae:
            raise ValueError(
                'No pre-trained VAE available. Please set "train_vae" to true in parameter "hyperparams" to train a VAE.'
            )

        path = os.path.expanduser(path)
        if not os.path.exists(path):
            os.makedirs(path)

        if self._vae_training:
            self._train_vae(path)

        # Authors say: 'For automatic explainer generation'
        flat_vae_bools = False
        cuda = torch.cuda.is_available()
        vae = VAE_gauss_cat_net(
            self._input_dimension,
            self._width,
            self._depth,
            self._latent_dim,
            pred_sig=False,
            lr=self._lr,
            cuda=cuda,
            flatten=flat_vae_bools,
        )

        vae.load(os.path.join(path, "theta_best.dat"))

        return vae

    def _train_vae(self, path):
        # training
        x_train, x_test = train_test_split(
            self._df_norm_enc_data.values, train_size=0.7
        )

        # Error message when training VAE using float 64: -> Change to: float 32
        # "Expected object of scalar type Float but got scalar type Double for argument #2 'mat1' in call to _th_addmm"
        x_train = np.float32(x_train)
        x_test = np.float32(x_test)

        training(
            x_train,
            x_test,
            self._input_dimension,
            path,
            self._width,
            self._depth,
            self._latent_dim,
            self._batch_size,
            self._epochs,
            self._lr,
            self._early_stop,
        )

    def get_counterfactuals(self, factuals: pd.DataFrame) -> pd.DataFrame:
        list_cfs = []

        # normalize and encode data and instance
        df_norm_enc_factual = self.encode_normalize_order_factuals(factuals)

        for index, row in df_norm_enc_factual.iterrows():
            counterfactual = vae_gradient_search(row.values, self._mlmodel, self._vae)
            list_cfs.append(counterfactual)

        # Convert output into correct format
        df_cfs = check_counterfactuals(self._mlmodel, list_cfs)

        return df_cfs



import pandas as pd


def success_rate(counterfactuals: pd.DataFrame) -> float:
    """
    Computes success rate for all counterfactuals

    Parameters
    ----------
    counterfactuals: All counterfactual examples inclusive nan values

    Returns
    -------

    """
    return (counterfactuals.dropna().shape[0]) / counterfactuals.shape[0]



import pandas as pd

from carla.recourse_methods.api import RecourseMethod
from carla.recourse_methods.autoencoder import CSVAE, train_variational_autoencoder
from carla.recourse_methods.catalog.crud.library import counterfactual_search
from carla.recourse_methods.processing import (
    check_counterfactuals,
    merge_default_parameters,
)


class CRUD(RecourseMethod):
    """
    Implementation of CRUD [1]_.

    Parameters
    ----------
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Dictionary containing hyperparameters. See notes below for its contents.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.
    Notes
    -----
    - Restriction
        * Currently working only with Pytorch models

    - Hyperparams
        Hyperparameter contains important information for the recourse method to initialize.
        Please make sure to pass all values as dict with the following keys.
        * "data_name": str
            name of the dataset
        * "lr": float, default: 0.008
            Learning rate for gradient descent.
        * "lambda_param": float, default: 0.001
            Weight factor for loss in counterfactual search.
        * "target_class" list, default: [0, 1]
            List of one-hot-encoded target class.
        * "binary_cat_features": bool, default: True
            If true, the encoding of x is done by drop_if_binary.
        * "max_iter": int, default: 2000
            Number of search steps to find a counterfactual.
        * "optimizer": {RMSprop, Adam}
            Optimizer for counterfactual search.
        * "vae_params": Dict
            With parameter for VAE.
            + "layers": list
                List with number of neurons per layer, incl. input and latent dimension.
            + "train": bool, default: True
                Decides if a new Autoencoder will be learned.
            + "epochs": int, default: 5
                Number of epcchs to train VAE
            + "lr": float, default: 1e-3
                Learning rate for VAE training
            + "batch_size": int, default: 32
                Batch-size for VAE training

    .. [1] M. Downs, J. Chu, Yacoby Y, Doshi-Velez F, WeiWei P. CRUDS: Counterfactual Recourse Using Disentangled
            Subspaces. ICML Workshop on Human Interpretability in Machine Learning. 2020 :1-23.
    """

    _DEFAULT_HYPERPARAMS = {
        "data_name": None,
        "target_class": [0, 1],
        "lambda_param": 0.001,
        "optimizer": "RMSprop",
        "lr": 0.008,
        "max_iter": 2000,
        "binary_cat_features": False,
        "vae_params": {
            "layers": None,
            "train": True,
            "epochs": 5,
            "lr": 1e-3,
            "batch_size": 32,
        },
    }

    def __init__(self, mlmodel, hyperparams):
        super().__init__(mlmodel)

        checked_hyperparams = merge_default_parameters(
            hyperparams, self._DEFAULT_HYPERPARAMS
        )

        self._target_class = checked_hyperparams["target_class"]
        self._lambda_param = checked_hyperparams["lambda_param"]
        self._optimizer = checked_hyperparams["optimizer"]
        self._lr = checked_hyperparams["lr"]
        self._max_iter = checked_hyperparams["max_iter"]
        self._binary_cat_features = checked_hyperparams["binary_cat_features"]

        df_enc_norm_data = self.encode_normalize_order_factuals(
            self._mlmodel.data.raw, with_target=True
        )

        vae_params = checked_hyperparams["vae_params"]
        self._vae = CSVAE(
            checked_hyperparams["data_name"],
            vae_params["layers"],
        )

        if vae_params["train"]:
            self._vae = train_variational_autoencoder(
                self._vae,
                self._mlmodel.data,
                self._mlmodel.scaler,
                self._mlmodel.encoder,
                self._mlmodel.feature_input_order,
                lambda_reg=None,
                epochs=vae_params["epochs"],
                lr=vae_params["lr"],
                batch_size=vae_params["batch_size"],
            )
        else:
            try:
                self._vae.load(df_enc_norm_data.shape[1] - 1)
            except FileNotFoundError as exc:
                raise FileNotFoundError(
                    "Loading of Autoencoder failed. {}".format(str(exc))
                )

    def get_counterfactuals(self, factuals: pd.DataFrame):
        df_enc_norm_fact = self.encode_normalize_order_factuals(
            factuals, with_target=True
        )

        # pay attention to categorical features
        encoded_feature_names = self._mlmodel.encoder.get_feature_names(
            self._mlmodel.data.categoricals
        )
        cat_features_indices = [
            df_enc_norm_fact.columns.get_loc(feature)
            for feature in encoded_feature_names
        ]

        df_cfs = df_enc_norm_fact.apply(
            lambda x: counterfactual_search(
                self._mlmodel,
                self._vae,
                x.reshape((1, -1)),
                cat_features_indices,
                self._binary_cat_features,
                self._target_class,
                self._lambda_param,
                self._optimizer,
                self._lr,
                self._max_iter,
            ),
            raw=True,
            axis=1,
        )

        cf_df = check_counterfactuals(
            self._mlmodel, df_cfs.drop(self._mlmodel.data.target, axis=1)
        )

        return cf_df



from abc import ABC, abstractmethod


class Data(ABC):
    """
    Abstract class to implement arbitrary datasets, which are provided by the user.
    """

    @property
    @abstractmethod
    def categoricals(self):
        """
        Provides the column names of categorical data.
        Column names do not contain encoded information as provided by a get_dummy() method (e.g., sex_female)

        Label name is not included.

        Returns
        -------
        list of Strings
            List of all categorical columns
        """
        pass

    @property
    @abstractmethod
    def continous(self):
        """
        Provides the column names of continuous data.

        Label name is not included.

        Returns
        -------
        list of Strings
            List of all continuous columns
        """
        pass

    @property
    @abstractmethod
    def immutables(self):
        """
        Provides the column names of immutable data.

        Label name is not included.

        Returns
        -------
        list of Strings
            List of all immutable columns
        """
        pass

    @property
    @abstractmethod
    def target(self):
        """
        Provides the name of the label column.

        Returns
        -------
        str
            Target label name
        """
        pass

    @property
    @abstractmethod
    def raw(self):
        """
        The raw Dataframe without encoding or normalization

        Returns
        -------
        pd.DataFrame
            Tabular data with raw information
        """
        pass



from typing import List

import numpy as np
import pandas as pd

from carla.models.api import MLModel


def redundancy(
    factuals: pd.DataFrame, counterfactuals: pd.DataFrame, mlmodel: MLModel
) -> List[List[int]]:
    """
    Computes Redundancy measure for every counterfactual

    Parameters
    ----------
    factuals: Encoded and normalized factual samples
    counterfactuals: Encoded and normalized counterfactual samples
    mlmodel: Black-box-model we want to discover

    Returns
    -------
    List with redundancy values per counterfactual sample
    """

    df_enc_norm_fact = factuals.reset_index(drop=True)
    df_cfs = counterfactuals.reset_index(drop=True)

    labels = df_cfs[mlmodel.data.target]
    df_cfs = df_cfs.drop(mlmodel.data.target, axis=1)
    df_cfs["redundancy"] = df_cfs.apply(
        lambda x: compute_redundancy(
            df_enc_norm_fact.iloc[x.name].values, x.values, mlmodel, labels.iloc[x.name]
        ),
        axis=1,
    )
    return df_cfs["redundancy"].values.reshape((-1, 1)).tolist()


def compute_redundancy(
    fact: np.ndarray, cf: np.ndarray, mlmodel: MLModel, label_value: int
) -> int:
    red = 0
    for col_idx in range(cf.shape[0]):  # input array has one-dimensional shape
        if fact[col_idx] != cf[col_idx]:
            temp_cf = np.copy(cf)

            temp_cf[col_idx] = fact[col_idx]

            temp_pred = np.argmax(mlmodel.predict_proba(temp_cf.reshape((1, -1))))

            if temp_pred == label_value:
                red += 1

    return red



from abc import ABC, abstractmethod

import pandas as pd

from carla.models.pipelining import encode, scale


class RecourseMethod(ABC):
    """
    Abstract class to implement custom recourse methods for a given black-box-model.

    Parameters
    ----------
    mlmodel: carla.models.MLModel
        Black-box-classifier we want to discover.

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Returns
    -------
    None
    """

    def __init__(self, mlmodel):
        self._mlmodel = mlmodel

    @abstractmethod
    def get_counterfactuals(self, factuals: pd.DataFrame):
        """
        Generate counterfactual examples for given factuals.

        Parameters
        ----------
        factuals: pd.DataFrame
            Not encoded and not normalised factual examples in two-dimensional shape (m, n).

        Returns
        -------
        pd.DataFrame
            Encoded and normalised counterfactual examples.
        """
        pass

    def encode_normalize_order_factuals(
        self, factuals: pd.DataFrame, with_target: bool = False
    ):
        """
        Uses encoder and scaler from black-box-model to preprocess data as needed.

        Parameters
        ----------
        factuals: pd.DataFrame
            Not encoded and not normalised factual examples in two-dimensional shape (m, n).
        with_target: bool, default: False
            If True, the output DataFrame contains label information

        Returns
        -------
        pd.DataFrame
            Encoded, normalized and ordered factual examples
        """
        # Prepare factuals
        querry_instances = factuals.copy()

        # check if querry_instances are not empty
        if querry_instances.shape[0] == 0:
            raise ValueError("Factuals should not be empty")

        factuals_enc_norm = scale(
            self._mlmodel.scaler, self._mlmodel.data.continous, querry_instances
        )
        factuals_enc_norm = encode(
            self._mlmodel.encoder, self._mlmodel.data.categoricals, factuals_enc_norm
        )

        label = [self._mlmodel.data.target] if with_target else []

        return factuals_enc_norm[self._mlmodel.feature_input_order + label]



from typing import Dict, Optional, Tuple

import numpy as np
import pandas as pd
import recourse as rs
from lime.lime_tabular import LimeTabularExplainer

from carla import log
from carla.recourse_methods.processing import encode_feature_names

from ...api import RecourseMethod
from ...processing.counterfactuals import merge_default_parameters


class ActionableRecourse(RecourseMethod):
    """
    Implementation of Actionable Recourse from Ustun et.al. [1]_

    Parameters
    ----------
    data : carla.data.Data
        Dataset
    mlmodel : carla.model.MLModel
        Black-Box-Model
    hyperparams : dict
        Dictionary containing hyperparameters. See Notes below to see its content.
    coeffs : np.ndArray, optional
        Coefficients. Will be approximated by LIME if None
    intercepts: np.ndArray, optional
        Intercepts. Will be approximated by LIME if None

    Methods
    -------
    get_counterfactuals:
        Generate counterfactual examples for given factuals.
    encode_normalize_order_factuals:
        Uses encoder and scaler from black-box-model to preprocess data as needed.

    Notes
    -----
    - Hyperparams
        Hyperparameter contains important information for the recourse method to initialize.
        Please make sure to pass all values as dict with the following keys.

        * "fs_size": int, default: 100
            Size of generated flipset.
        * "discretize": bool, default: False
            Parameter for LIME sampling.
        * "sample": boo, default: True
            Lime sampling around instance.
    - Restrictions
        *   Actionable Recourse (AR) supports only binary categorical features.
            See implementation at https://github.com/ustunb/actionable-recourse/blob/master/examples/ex_01_quickstart.ipynb
        *   AR is only defined on linear models. To make it work for arbitrary non-linear networks
            we need to find Actionab coefficients for every instance, for example with lime.
    - Warning
        *   AR does not always find a counterfactual example. The probability of finding one raises for a high size
            of flip set.

    .. [1] Berk Ustun, Alexander Spangher, and Y. Liu. 2019. Actionable Recourse in Linear Classification.
        InProceedings of the Conference on Fairness, Accountability, and Transparency (FAT*)
    """

    _DEFAULT_HYPERPARAMS = {
        "fs_size": 100,
        "discretize": False,
        "sample": True,
    }

    def __init__(
        self,
        mlmodel,
        hyperparams: Dict,
        coeffs: Optional[np.ndarray] = None,
        intercepts: Optional[np.ndarray] = None,
    ) -> None:
        super().__init__(mlmodel)
        self._data = mlmodel.data

        # normalize and encode data
        self._norm_enc_data = self.encode_normalize_order_factuals(
            self._data.raw, with_target=True
        )

        # Get hyperparameter
        checked_hyperparams = merge_default_parameters(
            hyperparams, self._DEFAULT_HYPERPARAMS
        )
        self._fs_size = checked_hyperparams["fs_size"]
        self._discretize_continuous = checked_hyperparams["discretize"]
        self._sample_around_instance = checked_hyperparams["sample"]

        # Build ActionSet
        self.action_set = rs.ActionSet(
            X=self._norm_enc_data[self._mlmodel.feature_input_order]
        )

        # transform immutable feature names into encoded feature names of self._data.encoded_normalized
        self._immutables = encode_feature_names(
            self._mlmodel.data.immutables, self._mlmodel.feature_input_order
        )

        for feature in self._immutables:
            self._action_set[feature].mutable = False
            self._action_set[feature].actionable = False

        self._coeffs, self._intercepts = coeffs, intercepts

    @property
    def action_set(self):
        """
        Contains dictionary with possible actions for every input feature.

        Returns
        -------
        dict
        """
        return self._action_set

    @action_set.setter
    def action_set(self, act_set):
        self._action_set = act_set

    def _get_lime_coefficients(
        self, factuals: pd.DataFrame
    ) -> Tuple[np.ndarray, np.ndarray]:
        """
        Actionable Recourse is only defined on linear models. To make it work for arbitrary non-linear networks
        we need to find the lime coefficients for every instance.

        Parameters
        ----------
        factuals : pd.DataFrame
            Instances we want to get lime coefficients

        Returns
        -------
        coeffs : np.ndArray
        intercepts : np.ndArray

        """
        coeffs = np.zeros(factuals.shape)
        intercepts = []
        lime_data = self._norm_enc_data[self._mlmodel.feature_input_order]
        lime_label = self._norm_enc_data[self._data.target]

        lime_exp = LimeTabularExplainer(
            training_data=lime_data.values,
            training_labels=lime_label,
            feature_names=self._mlmodel.feature_input_order,
            discretize_continuous=self._discretize_continuous,
            sample_around_instance=self._sample_around_instance,
            categorical_names=[
                cat
                for cat in self._mlmodel.feature_input_order
                if cat not in self._data.continous
            ]
            # self._data.encoded_normalized's categorical features contain feature name and value, separated by '_'
            # while self._data.categoricals do not contain those additional values.
        )

        for index, row in factuals.iterrows():
            factual = row.values
            explanations = lime_exp.explain_instance(
                factual,
                self._mlmodel.predict_proba,
                num_features=len(self._mlmodel.feature_input_order),
            )
            intercepts.append(explanations.intercept[1])

            for tpl in explanations.local_exp[1]:
                coeffs[index][tpl[0]] = tpl[1]

        return coeffs, np.array(intercepts)

    def get_counterfactuals(self, factuals: pd.DataFrame) -> pd.DataFrame:
        cfs = []
        coeffs = self._coeffs
        intercepts = self._intercepts
        action_set = self.action_set

        # to keep matching indexes for iterrows and coeffs
        factuals = factuals.reset_index()
        factuals_enc_norm = self.encode_normalize_order_factuals(factuals)

        # Check if we need lime to build coefficients
        if (coeffs is None) and (intercepts is None):
            log.info("Start generating LIME coefficients")
            coeffs, intercepts = self._get_lime_coefficients(factuals_enc_norm)
            log.info("Finished generating LIME coefficients")
        else:
            # Local explanations via LIME generate coeffs and intercepts per instance, while global explanations
            # via input parameter need to be set into correct shape [num_of_instances, num_of_features]
            coeffs = np.vstack([self._coeffs] * factuals.shape[0])
            intercepts = np.vstack([self._intercepts] * factuals.shape[0]).squeeze(
                axis=1
            )

        # generate counterfactuals
        for index, row in factuals_enc_norm.iterrows():
            # asserts are essential for mypy typechecking
            assert coeffs is not None
            assert intercepts is not None
            factual_enc_norm = row.values
            coeff = coeffs[index]
            intercept = intercepts[index]

            # Default counterfactual value if no action flips the prediction
            target_shape = factual_enc_norm.shape[0]
            empty = np.empty(target_shape)
            empty[:] = np.nan
            counterfactual = empty

            # Align action set to coefficients
            action_set.set_alignment(coefficients=coeff)

            # Build AR flipset
            fs = rs.Flipset(
                x=factual_enc_norm,
                action_set=action_set,
                coefficients=coeff,
                intercept=intercept,
            )
            try:
                fs_pop = fs.populate(total_items=self._fs_size)
            except (ValueError, KeyError):
                log.warning(
                    "Actionable Recourse is not able to produce a counterfactual explanation for instance {}".format(
                        index
                    )
                )
                log.warning(row.values)
                cfs.append(counterfactual)
                continue

            # Get actions to flip predictions
            actions = fs_pop.actions

            for action in actions:
                candidate_cf = (factual_enc_norm + action).reshape(
                    (1, -1)
                )  # Reshape to keep two-dim. input
                # Check if candidate counterfactual really flipps the prediction of ML model
                pred_cf = np.argmax(self._mlmodel.predict_proba(candidate_cf))
                pred_f = np.argmax(
                    self._mlmodel.predict_proba(factual_enc_norm.reshape((1, -1)))
                )
                if pred_cf != pred_f:
                    counterfactual = candidate_cf.squeeze()
                    break

            cfs.append(counterfactual)

        # Convert output into correct format
        cfs = np.array(cfs)
        df_cfs = pd.DataFrame(cfs, columns=self._mlmodel.feature_input_order)
        df_cfs[self._mlmodel.data.target] = np.argmax(
            self._mlmodel.predict_proba(cfs), axis=1
        )

        return df_cfs



# flake8: noqa

from .dataloader import VAEDataset
from .models import CSVAE, Autoencoder, VariationalAutoencoder
from .save_load import get_home
from .training import train_autoencoder, train_variational_autoencoder



# flake8: noqa

from .actionable_recourse import ActionableRecourse
from .cchvae import CCHVAE
from .cem import CEM
from .clue import Clue
from .crud import CRUD
from .dice import Dice
from .face import Face
from .growing_spheres import GrowingSpheres
from .revise import Revise
from .wachter import Wachter



from typing import Any, Callable, List, Tuple, Union

import numpy as np
import pandas as pd
import tensorflow as tf
import torch

from carla.data.catalog import DataCatalog
from carla.data.load_catalog import load_catalog
from carla.models.api import MLModel
from carla.models.pipelining import encode, order_data, scale

from .load_model import load_model


class MLModelCatalog(MLModel):
    """
    Use pretrained classifier.

    Parameters
    ----------
    data : data.catalog.DataCatalog Class
        Correct dataset for ML model.
    model_type : {'ann', 'linear'}
        Architecture.
    backend : {'tensorflow', 'pytorch'}
        Specifies the used framework.
    cache : boolean, default: True
        If True, try to load from the local cache first, and save to the cache.
        if a download is required.
    models_home : string, optional
        The directory in which to cache data; see :func:`get_models_home`.
    kws : keys and values, optional
        Additional keyword arguments are passed to passed through to the read model function
    use_pipeline : bool, default: False
        If true, the model uses a pipeline before predict and predict_proba to preprocess the input data.

    Methods
    -------
    predict:
        One-dimensional prediction of ml model for an output interval of [0, 1].
    predict_proba:
        Two-dimensional probability prediction of ml model
    get_pipeline_element:
        Returns a specific element of the pipeline
    perform_pipeline:
        Transforms input for prediction into correct form.

    Returns
    -------
    None
    """

    def __init__(
        self,
        data: DataCatalog,
        model_type: str,
        backend: str = "tensorflow",
        cache: bool = True,
        models_home: str = None,
        use_pipeline: bool = False,
        **kws
    ) -> None:
        """
        Constructor for pretrained ML models from the catalog.

        Possible backends are currently "pytorch" and "tensorflow".
        Possible models are corrently "ann".


        """
        self._backend = backend

        if self._backend == "pytorch":
            ext = "pt"
            encoding_method = "OneHot"
        elif self._backend == "tensorflow":
            ext = "h5"
            encoding_method = "OneHot_drop_binary"
        else:
            raise ValueError(
                "Backend not available, please choose between pytorch and tensorflow"
            )
        super().__init__(data, encoding_method=encoding_method)

        # Load catalog
        catalog_content = ["ann", "linear"]
        catalog = load_catalog("mlmodel_catalog.yaml", data.name, catalog_content)  # type: ignore

        if model_type not in catalog:
            raise ValueError("Model type not in model catalog")
        self._catalog = catalog[model_type][self._backend]
        self._feature_input_order = self._catalog["feature_order"]

        self._model = load_model(model_type, data.name, ext, cache, models_home, **kws)

        self._continuous = data.continous
        self._categoricals = data.categoricals

        # Preparing pipeline components
        self._use_pipeline = use_pipeline
        self._pipeline = self.__init_pipeline()

    def __init_pipeline(self) -> List[Tuple[str, Callable]]:
        return [
            ("scaler", lambda x: scale(self.scaler, self._continuous, x)),
            ("encoder", lambda x: encode(self.encoder, self._categoricals, x)),
            ("order", lambda x: order_data(self._feature_input_order, x)),
        ]

    def get_pipeline_element(self, key: str) -> Callable:
        """
        Returns a specific element of the pipeline

        Parameters
        ----------
        key : str
            Element of the pipeline we want to return

        Returns
        -------
        Pipeline element
        """
        key_idx = list(zip(*self._pipeline))[0].index(key)  # find key in pipeline
        return self._pipeline[key_idx][1]

    @property
    def pipeline(self) -> List[Tuple[str, Callable]]:
        """
        Returns transformations steps for input before predictions.

        Returns
        -------
        pipeline : list
            List of (name, transform) tuples that are chained in the order in which they are preformed.
        """
        return self._pipeline

    def perform_pipeline(self, df: pd.DataFrame) -> pd.DataFrame:
        """
        Transforms input for prediction into correct form.
        Only possible for DataFrames without preprocessing steps.

        Recommended to use to keep correct encodings, normalization and input order

        Parameters
        ----------
        df : pd.DataFrame
            Contains unnormalized and not encoded data.

        Returns
        -------
        output : pd.DataFrame
            Prediction input in correct order, normalized and encoded

        """
        output = df.copy()

        for trans_name, trans_function in self._pipeline:
            output = trans_function(output)

        return output

    @property
    def feature_input_order(self) -> List[str]:
        """
        Saves the required order of feature as list.

        Prevents confusion about correct order of input features in evaluation

        Returns
        -------
        ordered_features : list of str
            Correct order of input features for ml model
        """
        return self._feature_input_order

    @property
    def backend(self) -> str:
        """
        Describes the type of backend which is used for the ml model.

        E.g., tensorflow, pytorch, sklearn, ...

        Returns
        -------
        backend : str
            Used framework
        """
        return self._backend

    @property
    def raw_model(self) -> Any:
        """
        Returns the raw ml model built on its framework

        Returns
        -------
        ml_model : tensorflow, pytorch, sklearn model type
            Loaded model
        """
        return self._model

    def predict(
        self, x: Union[np.ndarray, pd.DataFrame, torch.Tensor, tf.Tensor]
    ) -> Union[np.ndarray, pd.DataFrame, torch.Tensor, tf.Tensor]:
        """
        One-dimensional prediction of ml model for an output interval of [0, 1]

        Shape of input dimension has to be always two-dimensional (e.g., (1, m), (n, m))

        Parameters
        ----------
        x : np.Array, pd.DataFrame, or backend specific (tensorflow or pytorch tensor)
            Tabular data of shape N x M (N number of instances, M number of features)

        Returns
        -------
        output : np.ndarray, or backend specific (tensorflow or pytorch tensor)
            Ml model prediction for interval [0, 1] with shape N x 1
        """

        if len(x.shape) != 2:
            raise ValueError("Input shape has to be two-dimensional")

        input = self.perform_pipeline(x) if self._use_pipeline else x

        if self._backend == "pytorch":
            return self.predict_proba(input)[:, 1].reshape((-1, 1))
        elif self._backend == "tensorflow":
            # keep output in shape N x 1
            return self._model.predict(input)[:, 1].reshape((-1, 1))
        else:
            raise ValueError(
                'Uncorrect backend value. Please use only "pytorch" or "tensorflow".'
            )

    def predict_proba(
        self, x: Union[np.ndarray, pd.DataFrame, torch.Tensor, tf.Tensor]
    ) -> Union[np.ndarray, pd.DataFrame, torch.Tensor, tf.Tensor]:
        """
        Two-dimensional probability prediction of ml model

        Shape of input dimension has to be always two-dimensional (e.g., (1, m), (n, m))

        Parameters
        ----------
        x : np.Array, pd.DataFrame, or backend specific (tensorflow or pytorch tensor)
            Tabular data of shape N x M (N number of instances, M number of features)

        Returns
        -------
        output : np.ndarray, or backend specific (tensorflow or pytorch tensor)
            Ml model prediction with shape N x 2
        """

        if len(x.shape) != 2:
            raise ValueError("Input shape has to be two-dimensional")

        input = self.perform_pipeline(x) if self._use_pipeline else x

        if self._backend == "pytorch":
            # Keep model and input on the same device
            device = "cuda" if torch.cuda.is_available() else "cpu"
            self._model = self._model.to(device)

            if isinstance(input, pd.DataFrame):
                input = input.values
            input, tensor_output = (
                (torch.Tensor(input), False)
                if not torch.is_tensor(input)
                else (input, True)
            )
            input = input.to(device)

            output = self._model(input)

            if tensor_output:
                return output
            else:
                return output.detach().cpu().numpy()

        elif self._backend == "tensorflow":
            return self._model.predict(input)
        else:
            raise ValueError(
                'Uncorrect backend value. Please use only "pytorch" or "tensorflow".'
            )

    @property
    def use_pipeline(self) -> bool:
        """
        Returns if the ML model uses the pipeline for predictions

        Returns
        -------
        bool
        """
        return self._use_pipeline

    @use_pipeline.setter
    def use_pipeline(self, use_pipe: bool) -> None:
        """
        Sets if the ML model should use the pipeline before prediction.

        Parameters
        ----------
        use_pipe : bool
            If true, the model uses a transformation pipeline before prediction.

        Returns
        -------

        """
        self._use_pipeline = use_pipe



import numpy as np
import torch
import torch.distributions as dists
from keras import backend as K
from torch import nn


def binary_crossentropy(y_true: np.ndarray, y_pred: np.ndarray) -> np.ndarray:
    return K.sum(K.binary_crossentropy(y_true, y_pred), axis=-1)


def csvae_loss(csvae, x_train, y_train):
    x = x_train.clone()
    x = x.float()
    y = y_train.clone()
    y = y.float()

    (
        x_mu,
        x_logvar,
        zw,
        y_pred,
        w_mu_encoder,
        w_logvar_encoder,
        w_mu_prior,
        w_logvar_prior,
        z_mu,
        z_logvar,
    ) = csvae.forward(x, y)

    x_recon = nn.MSELoss()(x_mu, x)

    w_dist = dists.MultivariateNormal(
        w_mu_encoder.flatten(), torch.diag(w_logvar_encoder.flatten().exp())
    )
    w_prior = dists.MultivariateNormal(
        w_mu_prior.flatten(), torch.diag(w_logvar_prior.flatten().exp())
    )
    w_kl = dists.kl.kl_divergence(w_dist, w_prior)

    z_dist = dists.MultivariateNormal(
        z_mu.flatten(), torch.diag(z_logvar.flatten().exp())
    )
    z_prior = dists.MultivariateNormal(
        torch.zeros(csvae.z_dim * z_mu.size()[0]),
        torch.eye(csvae.z_dim * z_mu.size()[0]),
    )
    z_kl = dists.kl.kl_divergence(z_dist, z_prior)

    y_pred_negentropy = (
        y_pred.log() * y_pred + (1 - y_pred).log() * (1 - y_pred)
    ).mean()

    class_label = torch.argmax(y, dim=1)
    y_recon = (
        100.0
        * torch.where(
            class_label == 1, -torch.log(y_pred[:, 1]), -torch.log(y_pred[:, 0])
        )
    ).mean()

    ELBO = 40 * x_recon + 0.2 * z_kl + 1 * w_kl + 110 * y_pred_negentropy

    return ELBO, x_recon, w_kl, z_kl, y_pred_negentropy, y_recon



import os
from typing import Optional
from urllib.error import HTTPError
from urllib.request import urlretrieve

import tensorflow as tf
import torch


def load_model(
    name: str,
    dataset: str,
    ext: str = "h5",
    cache: bool = True,
    models_home: Optional[str] = None,
    **kws,
):
    """Load an pretrained model from the online repository (requires internet).

    This function provides quick access to a number of models trained on example
    datasets that are commonly useful for evaluating counterfatual methods.

    Note that the models have been trained on the example datasets hosted on
    https://github.com/carla-recourse/cf-models.


    Use :func:`get_model_names` to see a list of available models given the dataset.

    Parameters
    ----------
    name : str
        Name of the model ``{name}.{ext}`` on https://github.com/carla-recourse/cf-models.
    dataset : str
        Name of the dataset the model has been trained on.
    cache : boolean, optional
        If True, try to load from the local cache first, and save to the cache
        if a download is required.
    models_home : string, optional
        The directory in which to cache data; see :func:`get_models_home`.
    kws : keys and values, optional
        Additional keyword arguments are passed to passed through to the read model function
    Returns
    -------
    model :  Tensorflow or PyTorch model
    """
    full_path = (
        "https://raw.githubusercontent.com/"
        "carla-recourse/cf-models/main/models/"
        f"{dataset}/{name}.{ext}"
    )

    if cache:
        cache_path = os.path.join(
            get_models_home(models_home), dataset, os.path.basename(full_path)
        )

        if not os.path.exists(cache_path):
            os.makedirs(os.path.dirname(cache_path), exist_ok=True)
            try:
                urlretrieve(full_path, cache_path)
            except HTTPError as e:
                raise ValueError(
                    f"'{name}' is not an available model for dataset '{dataset}'.", e
                )

        full_path = cache_path

    if ext == "pt":
        device = "cuda" if torch.cuda.is_available() else "cpu"
        model = torch.jit.load(
            full_path,
            map_location=device,
        )
        model = model.eval()
    elif ext == "h5":
        model = tf.keras.models.load_model(full_path, compile=False)
    else:
        raise NotImplementedError("Extension not supported:", ext)

    return model


def get_models_home(models_home=None):
    """Return a path to the cache directory for example models.

    This directory is then used by :func:`load_model`.

    If the ``models_home`` argument is not specified, it tries to read from the
    ``CF_MODELS`` environment variable and defaults to ``~/cf-bechmark/models``.

    """

    if models_home is None:
        models_home = os.environ.get("CF_MODELS", os.path.join("~", "carla", "models"))

    models_home = os.path.expanduser(models_home)
    if not os.path.exists(models_home):
        os.makedirs(models_home)

    return models_home



# flake8: noqa

from .counterfactuals import (
    check_counterfactuals,
    get_drop_columns_binary,
    merge_default_parameters,
    reconstruct_encoding_constraints,
)
from .immutables import encode_feature_names



import numpy as np
import pandas as pd
from numpy import linalg as LA

from carla import log


def hyper_sphere_coordindates(n_search_samples, instance, high, low, p_norm=2):

    # Implementation follows the Random Point Picking over a sphere
    # The algorithm's implementation follows: Pawelczyk, Broelemann & Kascneci (2020);
    # "Learning Counterfactual Explanations for Tabular Data" -- The Web Conference 2020 (WWW)
    # It ensures that points are sampled uniformly at random using insights from:
    # http://mathworld.wolfram.com/HyperspherePointPicking.html

    # This one implements the growing spheres method from
    # Thibaut Laugel et al (2018), "Comparison-based Inverse Classification for
    # Interpretability in Machine Learning" -- International Conference on Information Processing
    # and Management of Uncertainty in Knowledge-Based Systems (2018)

    """
    :param n_search_samples: int > 0
    :param instance: numpy input point array
    :param high: float>= 0, h>l; upper bound
    :param low: float>= 0, l<h; lower bound
    :param p: float>= 1; norm
    :return: candidate counterfactuals & distances
    """

    delta_instance = np.random.randn(n_search_samples, instance.shape[1])
    dist = np.random.rand(n_search_samples) * (high - low) + low  # length range [l, h)
    norm_p = LA.norm(delta_instance, ord=p_norm, axis=1)
    d_norm = np.divide(dist, norm_p).reshape(-1, 1)  # rescale/normalize factor
    delta_instance = np.multiply(delta_instance, d_norm)
    candidate_counterfactuals = instance + delta_instance

    return candidate_counterfactuals, dist


def growing_spheres_search(
    instance,
    keys_mutable,
    keys_immutable,
    continuous_cols,
    binary_cols,
    feature_order,
    model,
    n_search_samples=1000,
    p_norm=2,
    step=0.2,
    max_iter=1000,
):

    """
    :param instance: df
    :param step: float > 0; step_size for growing spheres
    :param n_search_samples: int > 0
    :param model: sklearn classifier object
    :param p_norm: float=>1; denotes the norm (classical: 1 or 2)
    :param max_iter: int > 0; maximum # iterations
    :param keys_mutable: list; list of input names we can search over
    :param keys_immutable: list; list of input names that may not be searched over
    :return:
    """  #

    # correct order of names
    keys_correct = feature_order
    # divide up keys
    keys_mutable_continuous = list(set(keys_mutable) - set(binary_cols))
    keys_mutable_binary = list(set(keys_mutable) - set(continuous_cols))

    # Divide data in 'mutable' and 'non-mutable'
    # In particular, divide data in 'mutable & binary' and 'mutable and continuous'
    instance_immutable_replicated = np.repeat(
        instance[keys_immutable].values.reshape(1, -1), n_search_samples, axis=0
    )
    instance_replicated = np.repeat(
        instance.values.reshape(1, -1), n_search_samples, axis=0
    )
    instance_mutable_replicated_continuous = np.repeat(
        instance[keys_mutable_continuous].values.reshape(1, -1),
        n_search_samples,
        axis=0,
    )
    # instance_mutable_replicated_binary = np.repeat(
    #     instance[keys_mutable_binary].values.reshape(1, -1), n_search_samples, axis=0
    # )

    # init step size for growing the sphere
    low = 0
    high = low + step

    # counter
    count = 0
    counter_step = 1

    # get predicted label of instance
    instance_label = np.argmax(model.predict_proba(instance.values.reshape(1, -1)))

    counterfactuals_found = False
    candidate_counterfactual_star = np.empty(
        instance_replicated.shape[1],
    )
    candidate_counterfactual_star[:] = np.nan
    while not counterfactuals_found or count > max_iter:
        count = count + counter_step

        # STEP 1 -- SAMPLE POINTS on hyper sphere around instance
        candidate_counterfactuals_continuous, _ = hyper_sphere_coordindates(
            n_search_samples, instance_mutable_replicated_continuous, high, low, p_norm
        )

        # sample random points from Bernoulli distribution
        candidate_counterfactuals_binary = np.random.binomial(
            n=1, p=0.5, size=n_search_samples * len(keys_mutable_binary)
        ).reshape(n_search_samples, -1)

        # make sure inputs are in correct order
        candidate_counterfactuals = pd.DataFrame(
            np.c_[
                instance_immutable_replicated,
                candidate_counterfactuals_continuous,
                candidate_counterfactuals_binary,
            ]
        )
        candidate_counterfactuals.columns = (
            keys_immutable + keys_mutable_continuous + keys_mutable_binary
        )
        # enforce correct order
        candidate_counterfactuals = candidate_counterfactuals[keys_correct]

        # STEP 2 -- COMPUTE l_1 DISTANCES
        if p_norm == 1:
            distances = np.abs(
                (candidate_counterfactuals.values - instance_replicated)
            ).sum(axis=1)
        elif p_norm == 2:
            distances = np.square(
                (candidate_counterfactuals.values - instance_replicated)
            ).sum(axis=1)
        else:
            raise ValueError("Distance not defined yet")

        # counterfactual labels
        y_candidate_logits = model.predict_proba(candidate_counterfactuals.values)
        y_candidate = np.argmax(y_candidate_logits, axis=1)
        indeces = np.where(y_candidate != instance_label)
        candidate_counterfactuals = candidate_counterfactuals.values[indeces]
        candidate_dist = distances[indeces]

        if len(candidate_dist) > 0:  # certain candidates generated
            min_index = np.argmin(candidate_dist)
            candidate_counterfactual_star = candidate_counterfactuals[min_index]
            counterfactuals_found = True

        # no candidate found & push search range outside
        low = high
        high = low + step

    return candidate_counterfactual_star



from __future__ import division

import torch

from carla.recourse_methods.catalog.clue.library.clue_ml.AE_models.AE.fc_gauss_cat import (
    VAE_gauss_cat_net,
)
from carla.recourse_methods.catalog.clue.library.clue_ml.AE_models.AE.train import (
    train_VAE,
)
from carla.recourse_methods.catalog.clue.library.clue_ml.src.utils import Datafeed


def training(
    x_train,
    x_test,
    input_dim_vec,
    path,
    width=10,
    depth=2,
    latent_dim=6,
    batch_size=128,
    nb_epochs=10,
    lr=1e-3,
    early_stop=25,
):
    """
    :param x_train: np.array; train set
    :param x_test: np.array; test set
    :param input_dim_vec: list; with dimensions of inputs (e.g. [1,2,1] for inputs ['#credit cards', 'sex' ,'age'])
    :param path: str; data set name (e.g. compas)
    :param width: int > 0; layer width
    :param depth: int > 0; layer depth
    :param latent_dim: int > 0; latent dim of VAE
    :param batch_size: int > 0; batch size
    :param nb_epochs: int > 0; # epochs
    :param lr: 0 < float < 1: learning rate
    :param early_stop: 0 < int < nb_epochs
    trains VAE and saves results
    """  #

    trainset = Datafeed(x_train, x_train, transform=None)
    valset = Datafeed(x_test, x_test, transform=None)

    # check whether GPU access
    cuda = torch.cuda.is_available()

    # load model architecture
    net = VAE_gauss_cat_net(
        input_dim_vec,
        width,
        depth,
        latent_dim,
        pred_sig=False,
        lr=lr,
        cuda=cuda,
        flatten=False,
    )

    # train & save model
    vlb_train, vlb_dev = train_VAE(
        net,
        path,
        batch_size,
        nb_epochs,
        trainset,
        valset,
        cuda=cuda,
        flat_ims=False,
        train_plot=False,
        early_stop=early_stop,
    )



# flake8: noqa

from .api import RecourseMethod
from .catalog import (
    CCHVAE,
    CEM,
    CRUD,
    ActionableRecourse,
    Clue,
    Dice,
    Face,
    GrowingSpheres,
    Revise,
    Wachter,
)



from typing import List


def encode_feature_names(imtbls: List[str], input_order: List[str]) -> List[str]:
    """
    Transforms not encoded immutable feature names into encoded ones

    Returns
    -------
    list
    """
    immutables = []
    # TODO: Maybe find a more elegant way to find encoded immutable feature names
    for feature in imtbls:
        if feature in input_order:
            immutables.append(feature)
        else:
            for cat in input_order:
                if cat not in immutables:
                    if feature in cat:
                        immutables.append(cat)
    return immutables
